"""Tests for flexmock ordered feature."""
# pylint: disable=missing-docstring,no-self-use,no-member
import re

from flexmock import exceptions, flexmock
from tests.utils import assert_raises


class OrderedTestCase:
    def test_ordered_on_different_methods(self):
        class String(str):
            pass

        string = String("abc")
        flexmock(string)
        string.should_call("startswith").with_args("asdf", 0, 4).ordered()
        string.should_call("endswith").ordered()
        with assert_raises(
            exceptions.CallOrderError,
            re.compile(
                r'endswith\("c"\) called before startswith'
                # Argument names are displayed in PyPy
                r'\((prefix=)?"asdf", (start=)?0, (end=)?4\)'
            ),
        ):
            string.endswith("c")

    def test_flexmock_ordered_worked_after_default_stub(self):
        mocked = flexmock()
        mocked.should_receive("bar")
        mocked.should_receive("bar").with_args("a").ordered()
        mocked.should_receive("bar").with_args("b").ordered()
        with assert_raises(exceptions.CallOrderError, 'bar("b") called before bar("a")'):
            mocked.bar("b")

    def test_flexmock_ordered_works_with_same_args(self):
        mocked = flexmock()
        mocked.should_receive("bar").ordered().and_return(1)
        mocked.should_receive("bar").ordered().and_return(2)
        assert mocked.bar() == 1
        assert mocked.bar() == 2

    def test_flexmock_ordered_works_with_same_args_after_default_stub(self):
        mocked = flexmock()
        mocked.should_receive("bar").and_return(9)
        mocked.should_receive("bar").ordered().and_return(1)
        mocked.should_receive("bar").ordered().and_return(2)
        assert mocked.bar() == 1
        assert mocked.bar() == 2
        assert mocked.bar() == 9



"""Proxy class for testing proxied objects."""


class Proxy:
    """Proxy to another object."""

    # OBJECT PROXYING (PYTHON RECIPE)
    # https://code.activestate.com/recipes/496741-object-proxying/

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    _special_names = [
        "__abs__",
        "__add__",
        "__and__",
        "__call__",
        "__cmp__",
        "__coerce__",
        "__contains__",
        "__delitem__",
        "__delslice__",
        "__div__",
        "__divmod__",
        "__eq__",
        "__float__",
        "__floordiv__",
        "__ge__",
        "__getitem__",
        "__getslice__",
        "__gt__",
        "__hash__",
        "__hex__",
        "__iadd__",
        "__iand__",
        "__idiv__",
        "__idivmod__",
        "__ifloordiv__",
        "__ilshift__",
        "__imod__",
        "__imul__",
        "__int__",
        "__invert__",
        "__ior__",
        "__ipow__",
        "__irshift__",
        "__isub__",
        "__iter__",
        "__itruediv__",
        "__ixor__",
        "__le__",
        "__len__",
        "__long__",
        "__lshift__",
        "__lt__",
        "__mod__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__oct__",
        "__or__",
        "__pos__",
        "__pow__",
        "__radd__",
        "__rand__",
        "__rdiv__",
        "__rdivmod__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__reversed__",
        "__rfloorfiv__",
        "__rlshift__",
        "__rmod__",
        "__rmul__",
        "__ror__",
        "__rpow__",
        "__rrshift__",
        "__rshift__",
        "__rsub__",
        "__rtruediv__",
        "__rxor__",
        "__setitem__",
        "__setslice__",
        "__sub__",
        "__truediv__",
        "__xor__",
        "next",
    ]

    @classmethod
    def _create_class_proxy(cls, theclass):
        """Creates a proxy for the given class."""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type(f"{cls.__name__}({theclass.__name__})", (cls,), namespace)

    def __new__(cls, obj, *args, **kwargs):
        """Creates an proxy instance referencing `obj`.

        (obj, *args, **kwargs) are passed to this class' __init__, so deriving
        classes can define an __init__ method of their own.

        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins



"""Test flexmock with pytest."""
# pylint: disable=missing-docstring,redefined-outer-name,no-self-use
import unittest

import pytest

from flexmock import flexmock
from flexmock._api import flexmock_teardown
from flexmock.exceptions import MethodCallError
from tests.features import FlexmockTestCase
from tests.utils import assert_raises


def test_module_level_test_for_pytest():
    flexmock(foo="bar").should_receive("foo").once()
    with assert_raises(
        MethodCallError, "foo() expected to be called exactly 1 time, called 0 times"
    ):
        flexmock_teardown()


@pytest.fixture()
def runtest_hook_fixture():
    return flexmock(foo="bar").should_receive("foo").once().mock()


def test_runtest_hook_with_fixture_for_pytest(runtest_hook_fixture):
    runtest_hook_fixture.foo()


def test_flexmock_teardown_works_with_pytest_part1():
    flexmock().should_receive("method1").ordered()


def test_flexmock_teardown_works_with_pytest_part2():
    mock = flexmock().should_receive("method2").ordered().mock()
    # Raises CallOrderError if flexmock teardown is not automatically called
    # after test part 1 above
    mock.method2()


class TestForPytest(FlexmockTestCase):
    def test_class_level_test_for_pytest(self):
        flexmock(foo="bar").should_receive("foo").once()
        with assert_raises(
            MethodCallError, "foo() expected to be called exactly 1 time, called 0 times"
        ):
            flexmock_teardown()


class TestUnittestClass(unittest.TestCase):
    def test_unittest(self):
        mocked = flexmock(a=2)
        mocked.should_receive("a").once()
        with assert_raises(
            MethodCallError, "a() expected to be called exactly 1 time, called 0 times"
        ):
            flexmock_teardown()


class TestFailureOnException:
    def _setup_failing_expectation(self):
        flexmock(foo="bar").should_receive("foo").once()

    # Use xfail to ensure we process exceptions as returned by _hook_into_pytest

    @pytest.mark.xfail(raises=RuntimeError)
    def test_exception(self):
        raise RuntimeError("TEST ERROR")

    @pytest.mark.xfail(raises=MethodCallError)
    def test_flexmock_raises_if_no_exception(self):
        self._setup_failing_expectation()

    @pytest.mark.xfail(raises=RuntimeError)
    def test_flexmock_doesnt_override_existing_exception(self):
        self._setup_failing_expectation()
        raise RuntimeError("Flexmock shouldn't suppress this exception")

    @pytest.mark.xfail(raises=AssertionError)
    def test_flexmock_doesnt_override_assertion(self):
        self._setup_failing_expectation()
        assert False, "Flexmock shouldn't suppress this assertion"



"""Test integrations that work with unittest testcases."""
# pylint: disable=missing-docstring,no-self-use
import unittest

from tests.features import FlexmockTestCase


class TestGenericIntegration(FlexmockTestCase, unittest.TestCase):
    pass



"""Tests for flexmock mocking feature."""
# pylint: disable=missing-docstring,no-self-use,no-member,too-many-lines
import os
import random
import re
import sys

from flexmock import exceptions, flexmock
from flexmock._api import AT_LEAST, AT_MOST, EXACTLY, FlexmockContainer, Mock, flexmock_teardown
from tests import some_module
from tests.utils import assert_raises


class MockingTestCase:
    def test_print_expectation(self):
        mock = flexmock()
        expectation = mock.should_receive("foo")
        assert str(expectation) == "foo() -> ()"

    def test_flexmock_should_create_mock_object(self):
        mock = flexmock()
        assert isinstance(mock, Mock)

    def test_flexmock_should_create_mock_object_from_dict(self):
        mock = flexmock(foo="foo", bar="bar")
        assert mock.foo == "foo"
        assert mock.bar == "bar"

    def test_flexmock_should_add_expectations(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo")
        assert "method_foo" in [x._name for x in FlexmockContainer.flexmock_objects[mock]]

    def test_flexmock_should_return_value(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar")
        mock.should_receive("method_bar").and_return("value_baz")
        assert mock.method_foo() == "value_bar"
        assert mock.method_bar() == "value_baz"

    def test_calling_deprecated_properties_should_raise_exception(self):
        error_msg = "Calling once, twice, never, or mock without parentheses has been deprecated"
        mock = flexmock()
        mock.should_receive("foo").once  # pylint: disable=expression-not-assigned
        with assert_raises(exceptions.FlexmockError, error_msg):
            flexmock_teardown()
        mock.should_receive("foo").twice  # pylint: disable=expression-not-assigned
        with assert_raises(exceptions.FlexmockError, error_msg):
            flexmock_teardown()
        mock.should_receive("foo").never  # pylint: disable=expression-not-assigned
        with assert_raises(exceptions.FlexmockError, error_msg):
            flexmock_teardown()
        mock.should_receive("foo").mock  # pylint: disable=expression-not-assigned
        with assert_raises(exceptions.FlexmockError, error_msg):
            flexmock_teardown()

    def test_type_flexmock_with_unicode_string_in_should_receive(self):
        class Foo:
            def method(self):
                return "bar"

        flexmock(Foo).should_receive("method").and_return("mocked_bar")

        instance = Foo()
        assert instance.method() == "mocked_bar"

    def test_flexmock_should_accept_shortcuts_for_creating_mock_object(self):
        mock = flexmock(attr1="value 1", attr2=lambda: "returning 2")
        assert mock.attr1 == "value 1"
        assert mock.attr2() == "returning 2"

    def test_flexmock_should_accept_shortcuts_for_creating_expectations(self):
        class Foo:
            def method1(self):
                pass

            def method2(self):
                pass

        instance = Foo()
        flexmock(instance, method1="returning 1", method2="returning 2")
        assert instance.method1() == "returning 1"
        assert instance.method2() == "returning 2"
        assert instance.method2() == "returning 2"

    def test_flexmock_expectations_returns_all(self):
        mock = flexmock(name="temp")
        assert mock not in FlexmockContainer.flexmock_objects
        mock.should_receive("method_foo")
        mock.should_receive("method_bar")
        assert len(FlexmockContainer.flexmock_objects[mock]) == 2

    def test_flexmock_expectations_returns_named_expectation(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo")
        assert FlexmockContainer.get_flexmock_expectation(mock, "method_foo")._name == "method_foo"

    def test_flexmock_expectations_returns_none_if_not_found(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo")
        assert FlexmockContainer.get_flexmock_expectation(mock, "method_bar") is None

    def test_flexmock_should_check_parameters(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").with_args("bar").and_return(1)
        mock.should_receive("method_foo").with_args("baz").and_return(2)
        assert mock.method_foo("bar") == 1
        assert mock.method_foo("baz") == 2

    def test_flexmock_should_keep_track_of_calls(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").with_args("foo").and_return(0)
        mock.should_receive("method_foo").with_args("bar").and_return(1)
        mock.should_receive("method_foo").with_args("baz").and_return(2)
        mock.method_foo("bar")
        mock.method_foo("bar")
        mock.method_foo("baz")
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo", ("foo",))
        assert expectation._times_called == 0
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo", ("bar",))
        assert expectation._times_called == 2
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo", ("baz",))
        assert expectation._times_called == 1

    def test_flexmock_should_set_expectation_call_numbers(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").times(1)
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        with assert_raises(
            exceptions.MethodCallError,
            "method_foo() expected to be called exactly 1 time, called 0 times",
        ):
            expectation._verify()
        mock.method_foo()
        expectation._verify()

    def test_flexmock_should_check_raised_exceptions(self):
        mock = flexmock(name="temp")

        class FakeException(Exception):
            pass

        mock.should_receive("method_foo").and_raise(FakeException)
        with assert_raises(FakeException, ""):
            mock.method_foo()
        assert FlexmockContainer.get_flexmock_expectation(mock, "method_foo")._times_called == 1

    def test_flexmock_should_check_raised_exceptions_instance_with_args(self):
        mock = flexmock(name="temp")

        class FakeException(Exception):
            def __init__(self, arg, arg2):
                # pylint: disable=super-init-not-called
                pass

        mock.should_receive("method_foo").and_raise(FakeException(1, arg2=2))
        with assert_raises(FakeException, "1"):
            mock.method_foo()
        assert FlexmockContainer.get_flexmock_expectation(mock, "method_foo")._times_called == 1

    def test_flexmock_should_check_raised_exceptions_class_with_args(self):
        mock = flexmock(name="temp")

        class FakeException(Exception):
            def __init__(self, arg, arg2):
                # pylint: disable=super-init-not-called
                pass

        mock.should_receive("method_foo").and_raise(FakeException, 1, arg2=2)
        with assert_raises(FakeException, "1"):
            mock.method_foo()
        assert FlexmockContainer.get_flexmock_expectation(mock, "method_foo")._times_called == 1

    def test_flexmock_should_match_any_args_by_default(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("bar")
        mock.should_receive("method_foo").with_args("baz").and_return("baz")
        assert mock.method_foo() == "bar"
        assert mock.method_foo(1) == "bar"
        assert mock.method_foo("foo", "bar") == "bar"
        assert mock.method_foo("baz") == "baz"

    def test_flexmock_should_fail_to_match_exactly_no_args_when_calling_with_args(self):
        mock = flexmock()
        mock.should_receive("method_foo").with_args()
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method_foo did not match expectations:\n"
                '  Received call:\tmethod_foo("baz")\n'
                "  Expected call[1]:\tmethod_foo()"
            ),
        ):
            mock.method_foo("baz")

    def test_flexmock_should_match_exactly_no_args(self):
        class Foo:
            def method(self):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args().and_return("baz")
        assert instance.method() == "baz"

    def test_expectation_dot_mock_should_return_mock(self):
        mock = flexmock(name="temp")
        assert mock.should_receive("method_foo").mock() == mock

    def test_flexmock_should_create_partial_new_style_object_mock(self):
        class User:
            def __init__(self, name=None):
                self.name = name

            def get_name(self):
                return self.name

            def set_name(self, name):
                self.name = name

        user = User()
        flexmock(user)
        user.should_receive("get_name").and_return("john")
        user.set_name("mike")
        assert user.get_name() == "john"

    def test_flexmock_should_create_partial_old_style_object_mock(self):
        class User:
            def __init__(self, name=None):
                self.name = name

            def get_name(self):
                return self.name

            def set_name(self, name):
                self.name = name

        user = User()
        flexmock(user)
        user.should_receive("get_name").and_return("john")
        user.set_name("mike")
        assert user.get_name() == "john"

    def test_flexmock_should_create_partial_new_style_class_mock(self):
        class User:
            def __init__(self):
                pass

            def get_name(self):
                pass

        flexmock(User)
        User.should_receive("get_name").and_return("mike")
        user = User()
        assert user.get_name() == "mike"

    def test_flexmock_should_create_partial_old_style_class_mock(self):
        class User:
            def __init__(self):
                pass

            def get_name(self):
                pass

        flexmock(User)
        User.should_receive("get_name").and_return("mike")
        user = User()
        assert user.get_name() == "mike"

    def test_flexmock_should_match_expectations_against_builtin_classes(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").with_args(str).and_return("got a string")
        mock.should_receive("method_foo").with_args(int).and_return("got an int")
        assert mock.method_foo("string!") == "got a string"
        assert mock.method_foo(23) == "got an int"
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method_foo did not match expectations:\n"
                "  Received call:\tmethod_foo(2.0)\n"
                "  Expected call[1]:\tmethod_foo(<class 'int'>)\n"
                "  Expected call[2]:\tmethod_foo(<class 'str'>)"
            ),
        ):
            mock.method_foo(2.0)

    def test_with_args_should_work_with_builtin_c_methods(self):
        flexmock(sys.stdout).should_call("write")  # set fall-through
        flexmock(sys.stdout).should_receive("write").with_args("flexmock_builtin_test")
        sys.stdout.write("flexmock_builtin_test")

    def test_with_args_should_work_with_builtin_c_functions(self):
        mocked = flexmock(sys)
        mocked.should_receive("exit").with_args(1).once()
        mocked.exit(1)
        flexmock_teardown()
        flexmock(os).should_receive("remove").with_args("path").once()
        os.remove("path")

    def test_with_args_should_work_with_builtin_python_methods(self):
        flexmock(random).should_receive("randint").with_args(1, 10).once()
        random.randint(1, 10)

    def test_flexmock_should_match_expectations_against_user_defined_classes(self):
        mock = flexmock(name="temp")

        class Foo:
            pass

        mock.should_receive("method_foo").with_args(Foo).and_return("got a Foo")
        assert mock.method_foo(Foo()) == "got a Foo"
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method_foo did not match expectations:\n"
                "  Received call:\tmethod_foo(1)\n"
                "  Expected call[1]:\tmethod_foo(<class 'tests.features.mocking.MockingTestCase"
                ".test_flexmock_should_match_expectations_against_user_defined_classes"
                ".<locals>.Foo'>)"
            ),
        ):
            mock.method_foo(1)

    def test_flexmock_configures_global_mocks_dict(self):
        mock = flexmock(name="temp")
        assert mock not in FlexmockContainer.flexmock_objects
        mock.should_receive("method_foo")
        assert mock in FlexmockContainer.flexmock_objects
        assert len(FlexmockContainer.flexmock_objects[mock]) == 1

    def test_flexmock_teardown_verifies_mocks(self):
        mock = flexmock(name="temp")
        mock.should_receive("verify_expectations").times(1)
        with assert_raises(
            exceptions.MethodCallError,
            "verify_expectations() expected to be called exactly 1 time, called 0 times",
        ):
            flexmock_teardown()

    def test_flexmock_teardown_does_not_verify_stubs(self):
        mock = flexmock(name="temp")
        mock.should_receive("verify_expectations")
        flexmock_teardown()

    def test_flexmock_preserves_stubbed_object_methods_between_tests(self):
        class User:
            def get_name(self):
                return "mike"

        user = User()
        flexmock(user).should_receive("get_name").and_return("john")
        assert user.get_name() == "john"
        flexmock_teardown()
        assert user.get_name() == "mike"

    def test_flexmock_preserves_stubbed_class_methods_between_tests(self):
        class User:
            def get_name(self):
                return "mike"

        user = User()
        flexmock(User).should_receive("get_name").and_return("john")
        assert user.get_name() == "john"
        flexmock_teardown()
        assert user.get_name() == "mike"

    def test_flexmock_removes_new_stubs_from_objects_after_tests(self):
        class User:
            def get_name(self):
                pass

        user = User()
        saved = user.get_name
        flexmock(user).should_receive("get_name").and_return("john")
        assert saved != user.get_name
        assert user.get_name() == "john"
        flexmock_teardown()
        assert user.get_name == saved

    def test_flexmock_removes_new_stubs_from_classes_after_tests(self):
        class User:
            def get_name(self):
                pass

        user = User()
        saved = user.get_name
        flexmock(User).should_receive("get_name").and_return("john")
        assert saved != user.get_name
        assert user.get_name() == "john"
        flexmock_teardown()
        assert user.get_name == saved

    def test_flexmock_removes_stubs_from_multiple_objects_on_teardown(self):
        class User:
            def get_name(self):
                pass

        class Group:
            def get_name(self):
                pass

        user = User()
        group = Group()
        saved1 = user.get_name
        saved2 = group.get_name
        flexmock(user).should_receive("get_name").and_return("john").once()
        flexmock(group).should_receive("get_name").and_return("john").once()
        assert saved1 != user.get_name
        assert saved2 != group.get_name
        assert user.get_name() == "john"
        assert group.get_name() == "john"
        flexmock_teardown()
        assert user.get_name == saved1
        assert group.get_name == saved2

    def test_flexmock_removes_stubs_from_multiple_classes_on_teardown(self):
        class User:
            def get_name(self):
                pass

        class Group:
            def get_name(self):
                pass

        user = User()
        group = User()
        saved1 = user.get_name
        saved2 = group.get_name
        flexmock(User).should_receive("get_name").and_return("john")
        flexmock(Group).should_receive("get_name").and_return("john")
        assert saved1 != user.get_name
        assert saved2 != group.get_name
        assert user.get_name() == "john"
        assert group.get_name() == "john"
        flexmock_teardown()
        assert user.get_name == saved1
        assert group.get_name == saved2

    def test_flexmock_stubs_are_callable(self):
        stub = flexmock()
        mocked = flexmock()
        mocked.should_receive("create").twice()
        stub.tickets = mocked
        # stub.tickets should work with and without parentheses
        stub.tickets().create()
        stub.tickets.create()
        flexmock_teardown()

    def test_flexmock_respects_at_least_when_called_less_than_requested(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("bar").at_least().twice()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._modifier == AT_LEAST
        mock.method_foo()
        with assert_raises(
            exceptions.MethodCallError,
            "method_foo() expected to be called at least 2 times, called 1 time",
        ):
            flexmock_teardown()

    def test_flexmock_respects_at_least_when_called_requested_number(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").at_least().once()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._modifier == AT_LEAST
        mock.method_foo()
        flexmock_teardown()

    def test_flexmock_respects_at_least_when_called_more_than_requested(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").at_least().once()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._modifier == AT_LEAST
        mock.method_foo()
        mock.method_foo()
        flexmock_teardown()

    def test_flexmock_respects_at_most_when_called_less_than_requested(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("bar").at_most().twice()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._modifier == AT_MOST
        mock.method_foo()
        flexmock_teardown()

    def test_flexmock_respects_at_most_when_called_requested_number(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").at_most().once()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._modifier == AT_MOST
        mock.method_foo()
        flexmock_teardown()

    def test_flexmock_respects_at_most_when_called_more_than_requested(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").at_most().once()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._modifier == AT_MOST
        mock.method_foo()
        with assert_raises(
            exceptions.MethodCallError,
            "method_foo() expected to be called at most 1 time, called 2 times",
        ):
            mock.method_foo()

    def test_flexmock_treats_once_as_times_one(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").once()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._expected_calls[EXACTLY] == 1
        with assert_raises(
            exceptions.MethodCallError,
            "method_foo() expected to be called exactly 1 time, called 0 times",
        ):
            flexmock_teardown()

    def test_flexmock_treats_twice_as_times_two(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").twice().and_return("value_bar")
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._expected_calls[EXACTLY] == 2
        with assert_raises(
            exceptions.MethodCallError,
            "method_foo() expected to be called exactly 2 times, called 0 times",
        ):
            flexmock_teardown()

    def test_flexmock_works_with_never_when_true(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").never()
        expectation = FlexmockContainer.get_flexmock_expectation(mock, "method_foo")
        assert expectation._expected_calls[EXACTLY] == 0
        flexmock_teardown()

    def test_flexmock_works_with_never_when_false(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").and_return("value_bar").never()
        with assert_raises(
            exceptions.MethodCallError,
            "method_foo() expected to be called exactly 0 times, called 1 time",
        ):
            mock.method_foo()

    def test_flexmock_get_flexmock_expectation_should_work_with_args(self):
        mock = flexmock(name="temp")
        mock.should_receive("method_foo").with_args("value_bar")
        assert FlexmockContainer.get_flexmock_expectation(mock, "method_foo", "value_bar")

    def test_mock_kwargs_only_func_mock_all(self):
        flexmock(some_module).should_receive("kwargs_only_func1").with_args(
            1, bar=2, baz=3
        ).and_return(123)
        assert some_module.kwargs_only_func1(1, bar=2, baz=3) == 123

    def test_mock_kwargs_only_func_mock_required(self):
        flexmock(some_module).should_receive("kwargs_only_func1").with_args(1, bar=2).and_return(
            123
        )
        assert some_module.kwargs_only_func1(1, bar=2) == 123

    def test_mock_kwargs_only_func_fails_if_required_not_provided(self):
        with assert_raises(
            exceptions.MethodSignatureError,
            'kwargs_only_func1 requires keyword-only argument "bar"',
        ):
            flexmock(some_module).should_receive("kwargs_only_func1").with_args(1)
        with assert_raises(
            exceptions.MethodSignatureError,
            'kwargs_only_func2 requires keyword-only arguments "bar", "baz"',
        ):
            flexmock(some_module).should_receive("kwargs_only_func2").with_args(2)

    def test_context_manager_on_instance(self):
        class CM:
            def __enter__(self):
                pass

            def __exit__(self, *_):
                pass

        instance = CM()
        flexmock(instance).should_call("__enter__").once()
        flexmock(instance).should_call("__exit__").once()
        with instance:
            pass
        flexmock_teardown()

    def test_context_manager_on_class(self):
        class CM:
            def __enter__(self):
                pass

            def __exit__(self, *_):
                pass

        instance = CM()
        flexmock(CM).should_receive("__enter__").once()
        flexmock(CM).should_receive("__exit__").once()
        with instance:
            pass
        flexmock_teardown()

    def test_flexmock_should_support_with(self):
        mocked = flexmock()
        with mocked as mock:
            mock.should_receive("bar").and_return("baz")
        assert mocked.bar() == "baz"

    def test_builtin_open(self):
        mock = flexmock(sys.modules["builtins"])
        fake_fd = flexmock(read=lambda: "some data")
        mock.should_receive("open").once().with_args("file_name").and_return(fake_fd)
        with open("file_name") as file:  # pylint: disable=unspecified-encoding
            data = file.read()
        assert data == "some data"

    def test_mock_multiple_properties(self):
        class Foo:
            @property
            def prop1(self):
                return "prop1"

            @property
            def prop2(self):
                return "prop2"

        mocked = Foo()
        flexmock(mocked, prop1="mocked1", prop2="mocked2")
        assert mocked.prop1 == "mocked1"
        assert mocked.prop2 == "mocked2"

    def test_mock_property_with_attribute_on_instance(self):
        class Foo:
            @property
            def method(self):
                return "bar"

        foo1 = Foo()
        foo2 = Foo()
        foo3 = Foo()
        flexmock(foo1, method="baz")
        flexmock(foo2, method="baz2")
        assert foo1.method == "baz"
        assert foo2.method == "baz2"
        assert foo3.method == "bar"
        flexmock_teardown()
        assert hasattr(Foo, "_flexmock__method") is False, "Property method not cleaned up"
        assert foo1.method == "bar"
        assert foo2.method == "bar"
        assert foo3.method == "bar"

    def test_mock_property_with_attribute_on_class(self):
        class Foo:
            @property
            def method(self):
                return "bar"

        foo1 = Foo()
        foo2 = Foo()
        flexmock(Foo, method="baz")
        assert foo1.method == "baz"
        assert foo2.method == "baz"
        flexmock_teardown()
        assert hasattr(Foo, "_flexmock__method") is False, "Property method not cleaned up"
        assert foo1.method == "bar"
        assert foo2.method == "bar"

    def test_verifying_methods_when_mocking_module(self):
        # previously, we had problems with recognizing methods vs functions if the mocked
        # object was an imported module
        flexmock(some_module).should_receive("ModuleClass").with_args(1, 2)
        flexmock(some_module).should_receive("module_function").with_args(1, 2)

    def test_fake_object_supporting_iteration(self):
        mocked = flexmock()
        mocked.should_receive("__iter__").and_yield(1, 2, 3)
        assert list(mocked) == [1, 2, 3]

    def test_with_args_for_single_named_arg_with_optional_args(self):
        class FooClass:
            def method(self, one, two="optional"):
                pass

        flexmock(FooClass).should_receive("method").with_args(one=1)

    def test_with_args_doesnt_set_max_when_using_varargs(self):
        class FooClass:
            def method(self, *args):
                pass

        flexmock(FooClass).should_receive("method").with_args(1, 2, 3)

    def test_with_args_doesnt_set_max_when_using_kwargs(self):
        class FooClass:
            def method(self, **kwargs):
                pass

        flexmock(FooClass).should_receive("method").with_args(1, 2, 3)

    def test_with_args_blows_up_on_too_few_args(self):
        class FooClass:
            def method1(self, arg1):
                pass

            def method2(self, arg1, arg2, arg3=1):
                pass

        mock = flexmock(FooClass)
        expectation1 = mock.should_receive("method1")
        expectation2 = mock.should_receive("method2")
        with assert_raises(
            exceptions.MethodSignatureError,
            "method1 requires at least 1 argument, expectation provided 0",
        ):
            expectation1.with_args()
        with assert_raises(
            exceptions.MethodSignatureError,
            "method2 requires at least 2 arguments, expectation provided 1",
        ):
            expectation2.with_args(1)

    def test_with_args_blows_up_on_too_few_args_with_kwargs(self):
        class FooClass:
            def method(self, arg1, arg2, arg3=1):
                pass

        expectation = flexmock(FooClass).should_receive("method")
        with assert_raises(
            exceptions.MethodSignatureError,
            "method requires at least 3 arguments, expectation provided 2",
        ):
            expectation.with_args(1, arg3=2)

    def test_with_args_blows_up_on_too_many_args(self):
        class FooClass:
            def method(self, arg1, arg2, arg3=1):
                pass

        expectation = flexmock(FooClass).should_receive("method")
        with assert_raises(
            exceptions.MethodSignatureError,
            "method requires at most 3 arguments, expectation provided 4",
        ):
            expectation.with_args(1, 2, 3, 4)

    def test_with_args_blows_up_on_kwarg_overlapping_positional(self):
        class FooClass:
            def method(self, arg1, arg2, arg3=1, **kwargs):
                pass

        expectation = flexmock(FooClass).should_receive("method")
        with assert_raises(
            exceptions.MethodSignatureError,
            "['arg3'] already given as positional argument to method",
        ):
            expectation.with_args(1, 2, 3, arg3=2)
        with assert_raises(
            exceptions.MethodSignatureError,
            "['arg3', 'arg2'] already given as positional arguments to method",
        ):
            expectation.with_args(1, 2, 3, arg3=2, arg2=3)

    def test_with_args_blows_up_on_invalid_kwarg(self):
        class FooClass:
            def method(self, arg1, arg2, arg3=1):
                pass

        with assert_raises(
            exceptions.MethodSignatureError, "d is not a valid keyword argument to method"
        ):
            flexmock(FooClass).should_receive("method").with_args(1, 2, d=2)

    def test_with_args_ignores_invalid_args_on_flexmock_instances(self):
        instance = flexmock(bar=lambda x: x)
        instance.should_receive("bar").with_args("stuff")
        instance.bar("stuff")

    def test_with_args_does_not_compensate_for_self_on_static_instance_methods(self):
        class FooClass:
            @staticmethod
            def method(arg):
                pass

        instance = FooClass()
        flexmock(instance).should_receive("method").with_args("stuff")
        instance.method("stuff")

    def test_with_args_does_not_compensate_for_self_on_static_class_methods(self):
        class FooClass:
            @staticmethod
            def method(arg):
                pass

        flexmock(FooClass).should_receive("method").with_args("stuff")
        FooClass.method("stuff")

    def test_with_args_does_compensate_for_cls_on_class_methods(self):
        class FooClass:
            @classmethod
            def method(cls, arg):
                pass

        instance = FooClass()
        flexmock(instance).should_receive("method").with_args("stuff")
        instance.method("stuff")

    def test_calling_with_keyword_args_matches_mock_with_positional_args(self):
        class FooClass:
            def method(self, arg1, arg2, arg3):
                pass

        instance = FooClass()
        flexmock(instance).should_receive("method").with_args(1, 2, 3).once()
        instance.method(arg1=1, arg2=2, arg3=3)

    def test_calling_with_positional_args_matches_mock_with_kwargs(self):
        class FooClass:
            def method(self, arg1, arg2, arg3):
                pass

        instance = FooClass()
        flexmock(instance).should_receive("method").with_args(arg1=1, arg2=2, arg3=3).once()
        instance.method(1, 2, arg3=3)

    def test_and_return_defaults_to_none_with_no_arguments(self):
        mocked = flexmock()
        mocked.should_receive("bar").and_return()
        assert mocked.bar() is None

    def test_non_callable_attributes_fail_to_set_expectations(self):
        class FooClass:
            attribute = 1

        instance = FooClass()
        expectation = flexmock(instance).should_receive("attribute").and_return(2)
        with assert_raises(exceptions.FlexmockError, "can't use times() with attribute stubs"):
            expectation.times(1)
        with assert_raises(exceptions.FlexmockError, "can't use with_args() with attribute stubs"):
            expectation.with_args(())
        with assert_raises(
            exceptions.FlexmockError, "can't use replace_with() with attribute/property stubs"
        ):
            expectation.replace_with(lambda x: x)
        with assert_raises(exceptions.FlexmockError, "can't use and_raise() with attribute stubs"):
            expectation.and_raise(Exception)
        with assert_raises(exceptions.FlexmockError, "can't use when() with attribute stubs"):
            expectation.when(lambda x: x)
        with assert_raises(exceptions.FlexmockError, "can't use and_yield() with attribute stubs"):
            expectation.and_yield(1)
        with assert_raises(exceptions.FlexmockError, "can't use ordered() with attribute stubs"):
            object.__getattribute__(expectation, "ordered")()
        with assert_raises(exceptions.FlexmockError, "can't use at_least() with attribute stubs"):
            object.__getattribute__(expectation, "at_least")()
        with assert_raises(exceptions.FlexmockError, "can't use at_most() with attribute stubs"):
            object.__getattribute__(expectation, "at_most")()
        with assert_raises(exceptions.FlexmockError, "can't use one_by_one() with attribute stubs"):
            object.__getattribute__(expectation, "one_by_one")()

    def test_should_chain_attributes(self):
        class Class1:
            x = 1

        class Class2:
            class1 = Class1()

        class Class3:
            class2 = Class2()

        class3 = Class3()
        mocked = flexmock(class3)
        mocked.should_receive("class2.class1.x").and_return(2)
        assert mocked.class2.class1.x == 2
        flexmock_teardown()
        assert mocked.class2.class1.x == 1

    def test_mocking_subclass_of_str(self):
        class String(str):
            pass

        string = String()
        flexmock(string, endswith="fake")
        assert string.endswith("stuff") == "fake"
        flexmock_teardown()
        assert string.endswith("stuff") is False

    def test_proper_reset_of_subclass_methods(self):
        class ClassA:
            def method(self):
                return "a"

        class ClassB(ClassA):
            def method(self):
                return "b"

        flexmock(ClassB).should_receive("method").and_return("1")
        flexmock_teardown()
        assert ClassB().method() == "b"

    def test_flexmock_class_returns_same_object_on_repeated_calls(self):
        class Foo:
            pass

        first = flexmock(Foo)
        second = flexmock(Foo)
        assert first is second

    def test_flexmock_object_returns_same_object_on_repeated_calls(self):
        class Foo:
            pass

        instance = Foo()
        first = flexmock(instance)
        second = flexmock(instance)
        assert first is second

    def test_mocking_down_the_inheritance_chain_class_to_class(self):
        class Parent:
            def method1(self):
                pass

        class Child(Parent):
            def method2(self):
                pass

        flexmock(Parent).should_receive("method1").and_return("outer")
        flexmock(Child).should_receive("method2").and_return("inner")
        assert Parent().method1() == "outer"
        assert Child().method2() == "inner"

    def test_flexmock_should_mock_the_same_method_multiple_times(self):
        class Foo:
            def method(self):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").and_return(1)
        assert instance.method() == 1
        flexmock(instance).should_receive("method").and_return(2)
        assert instance.method() == 2
        flexmock(instance).should_receive("method").and_return(3)
        assert instance.method() == 3
        flexmock(instance).should_receive("method").and_return(4)
        assert instance.method() == 4

    def test_new_instances_should_be_a_method(self):
        class Foo:
            pass

        flexmock(Foo).new_instances("bar")
        assert Foo() == "bar"
        flexmock_teardown()
        assert Foo() != "bar"

    def test_new_instances_raises_error_when_not_a_class(self):
        class Foo:
            pass

        instance = Foo()
        flexmock(instance)
        with assert_raises(
            exceptions.FlexmockError, "new_instances can only be called on a class mock"
        ):
            instance.new_instances("bar")

    def test_new_instances_works_with_multiple_return_values(self):
        class Foo:
            pass

        flexmock(Foo).new_instances("foo", "bar")
        assert Foo() == "foo"
        assert Foo() == "bar"

    def test_should_receive_should_not_replace_flexmock_methods(self):
        class Foo:
            pass

        instance = Foo()
        flexmock(instance)
        for method_name in ["should_receive", "should_call", "new_instances"]:
            with assert_raises(exceptions.FlexmockError, "unable to replace flexmock methods"):
                instance.should_receive(method_name)

    def test_flexmock_should_not_add_methods_if_they_already_exist(self):
        class Foo:
            def should_receive(self):
                return "real"

            def method(self):
                pass

        instance = Foo()
        mock = flexmock(instance)
        assert instance.should_receive() == "real"
        assert "should_call" not in dir(instance)
        assert "new_instances" not in dir(instance)
        mock.should_receive("method").and_return("baz")
        assert instance.method() == "baz"
        flexmock_teardown()
        assert instance.should_receive() == "real"

    def test_flexmock_should_not_add_class_methods_if_they_already_exist(self):
        class Foo:
            def should_receive(self):
                return "real"

            def method(self):
                pass

        instance = Foo()
        mock = flexmock(Foo)
        assert instance.should_receive() == "real"
        assert "should_call" not in dir(Foo)
        assert "new_instances" not in dir(Foo)
        mock.should_receive("method").and_return("baz")
        assert instance.method() == "baz"
        flexmock_teardown()
        assert instance.should_receive() == "real"

    def test_flexmock_should_replace_method(self):
        class Foo:
            def method(self, arg):
                return arg

        instance = Foo()
        flexmock(instance).should_receive("method").replace_with(lambda x: x == 5)
        assert instance.method(5) is True
        assert instance.method(4) is False

    def test_flexmock_should_replace_cannot_be_specified_twice(self):
        class Foo:
            def method(self, arg):
                return arg

        instance = Foo()
        expectation = flexmock(instance).should_receive("method").replace_with(lambda x: x == 5)
        with assert_raises(exceptions.FlexmockError, "replace_with cannot be specified twice"):
            expectation.replace_with(lambda x: x == 3)

    def test_should_give_reasonable_error_for_builtins(self):
        with assert_raises(
            exceptions.MockBuiltinError,
            (
                "Python does not allow you to mock builtin objects or modules. "
                "Consider wrapping it in a class you can mock instead"
            ),
        ):
            flexmock(object)

    def test_should_give_reasonable_error_for_instances_of_builtins(self):
        with assert_raises(
            exceptions.MockBuiltinError,
            (
                "Python does not allow you to mock instances of builtin objects. "
                "Consider wrapping it in a class you can mock instead"
            ),
        ):
            flexmock(object())

    def test_mock_chained_method_calls_works_with_one_level(self):
        class Foo:
            def method2(self):
                return "foo"

        class Bar:
            def method1(self):
                return Foo()

        instance = Bar()
        assert instance.method1().method2() == "foo"
        flexmock(instance).should_receive("method1.method2").and_return("bar")
        assert instance.method1().method2() == "bar"

    def test_mock_chained_method_supports_args_and_mocks(self):
        class Foo:
            def method2(self, arg):
                return arg

        class Bar:
            def method1(self):
                return Foo()

        instance = Bar()
        assert instance.method1().method2("foo") == "foo"
        flexmock(instance).should_receive("method1.method2").with_args("foo").and_return(
            "bar"
        ).once()
        assert instance.method1().method2("foo") == "bar"
        flexmock_teardown()
        flexmock(instance).should_receive("method1.method2").with_args("foo").and_return(
            "bar"
        ).once()
        with assert_raises(
            exceptions.MethodCallError,
            'method2("foo") expected to be called exactly 1 time, called 0 times',
        ):
            flexmock_teardown()

    def test_mock_chained_method_calls_works_with_more_than_one_level(self):
        class Baz:
            def method3(self):
                return "foo"

        class Foo:
            def method2(self):
                return Baz()

        class Bar:
            def method1(self):
                return Foo()

        instance = Bar()
        assert instance.method1().method2().method3() == "foo"
        flexmock(instance).should_receive("method1.method2.method3").and_return("bar")
        assert instance.method1().method2().method3() == "bar"

    def test_flexmock_should_fail_mocking_nonexistent_methods(self):
        class User:
            pass

        user = User()
        with assert_raises(
            exceptions.FlexmockError,
            re.compile(
                r"<.+\.<locals>\.User object at 0x.+> does not have attribute 'nonexistent'"
            ),
        ):
            flexmock(user).should_receive("nonexistent")

    def test_new_instances_should_blow_up_on_should_receive(self):
        class User:
            pass

        mock = flexmock(User).new_instances(None).mock()
        with assert_raises(exceptions.FlexmockError, "User does not have attribute 'foo'"):
            mock.should_receive("foo")

    def test_flexmock_should_support_mocking_classes_as_functions(self):
        flexmock(some_module).should_receive("SomeClass").and_return("yay")
        assert some_module.SomeClass() == "yay"

    def test_flexmock_should_properly_restore_class_methods(self):
        class User:
            @classmethod
            def get_stuff(cls):
                return cls.__name__

        assert isinstance(User.__dict__["get_stuff"], classmethod)
        assert User.get_stuff() == "User"
        flexmock(User).should_receive("get_stuff").and_return("foo")
        assert User.get_stuff() == "foo"
        flexmock_teardown()
        assert User.get_stuff() == "User"
        assert isinstance(User.__dict__["get_stuff"], classmethod)

    def test_flexmock_should_mock_same_class_twice(self):
        class Foo:
            pass

        flexmock(Foo)
        flexmock(Foo)

    def test_and_raise_with_invalid_arguments(self):
        with assert_raises(
            exceptions.FlexmockError,
            "can't initialize <class 'Exception'> with the given arguments",
        ):
            flexmock().should_receive("foo").and_raise(Exception, 1, bar=2)

    def test_flexmock_should_mock_iter_on_new_style_instances(self):
        class Foo:
            def __iter__(self):
                yield None

        old = Foo.__iter__
        foo1 = Foo()
        foo2 = Foo()
        foo3 = Foo()
        flexmock(foo1, __iter__=iter([1, 2, 3]))
        flexmock(foo2, __iter__=iter([3, 4, 5]))
        assert list(foo1) == [1, 2, 3]
        assert list(foo2) == [3, 4, 5]
        assert list(foo3) == [None]
        assert foo1.__iter__ is not old
        assert foo2.__iter__ is not old
        assert foo3.__iter__ is not old
        flexmock_teardown()
        assert list(foo1) == [None]
        assert list(foo2) == [None]
        assert list(foo3) == [None]
        assert Foo.__iter__ is old, f"{Foo.__iter__} != {old}"

    def test_flexmock_should_mock_private_methods_with_leading_underscores(self):
        class ClassWithPrivateMethods:
            def __private_instance_method(self):
                pass

            @classmethod
            def __private_class_method(cls):
                pass

            def instance_method(self):
                return self.__private_instance_method()

            @classmethod
            def class_method(cls):
                return cls.__private_class_method()

        # Instance
        instance = ClassWithPrivateMethods()
        flexmock(instance).should_receive("__private_instance_method").and_return("bar")
        assert instance.instance_method() == "bar"

        # Class
        flexmock(ClassWithPrivateMethods).should_receive("__private_class_method").and_return("baz")
        assert ClassWithPrivateMethods.class_method() == "baz"

    def test_flexmock_should_mock_generators(self):
        class Gen:
            def method(self):
                pass

        gen = Gen()
        flexmock(gen).should_receive("method").and_yield(*range(1, 10))
        output = list(gen.method())
        assert list(range(1, 10)) == output

    def test_flexmock_should_mock_private_methods(self):
        class Foo:
            def __private_method(self):
                return "foo"

            def public_method(self):
                return self.__private_method()

        instance = Foo()
        flexmock(instance).should_receive("__private_method").and_return("bar")
        assert instance.public_method() == "bar"

    def test_flexmock_should_mock_special_methods(self):
        class Foo:
            def __special_method__(self):
                return "foo"

            def public_method(self):
                return self.__special_method__()

        instance = Foo()
        flexmock(instance).should_receive("__special_method__").and_return("bar")
        assert instance.public_method() == "bar"

    def test_flexmock_should_mock_double_underscore_method(self):
        class Foo:
            def __(self):
                return "foo"

            def public_method(self):
                return self.__()

        instance = Foo()
        flexmock(instance).should_receive("__").and_return("bar")
        assert instance.public_method() == "bar"

    def test_flexmock_should_mock_private_class_methods(self):
        class Foo:
            def __iter__(self):
                pass

        flexmock(Foo).should_receive("__iter__").and_yield(1, 2, 3)
        assert list(Foo()) == [1, 2, 3]

    def test_flexmock_doesnt_error_on_properly_ordered_expectations(self):
        class Foo:
            def method1(self):
                pass

            def method2(self, arg1):
                pass

            def method3(self):
                pass

            def method4(self):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method1")
        flexmock(instance).should_receive("method2").with_args("a").ordered()
        flexmock(instance).should_receive("method3")
        flexmock(instance).should_receive("method2").with_args("b").ordered()
        flexmock(instance).should_receive("method4")
        instance.method3()
        instance.method2("a")
        instance.method2("b")
        instance.method4()
        instance.method1()

    def test_flexmock_errors_on_improperly_ordered_expectations(self):
        class Foo:
            def method1(self, arg1):
                pass

        instance = Foo()
        flexmock(instance)
        instance.should_receive("method1").with_args("a").ordered()
        instance.should_receive("method1").with_args("b").ordered()
        with assert_raises(
            exceptions.CallOrderError, 'method1("b") called before method1(arg1="a")'
        ):
            instance.method1("b")

    def test_flexmock_should_accept_multiple_return_values(self):
        class Foo:
            def method1(self):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method1").and_return(1, 5).and_return(2)
        assert instance.method1() == (1, 5)
        assert instance.method1() == 2
        assert instance.method1() == (1, 5)
        assert instance.method1() == 2

    def test_flexmock_should_accept_multiple_return_values_with_shortcut(self):
        class Foo:
            def method1(self):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method1").and_return(1, 2).one_by_one()
        assert instance.method1() == 1
        assert instance.method1() == 2
        assert instance.method1() == 1
        assert instance.method1() == 2

    def test_flexmock_should_accept_multiple_return_values_with_one_by_one(self):
        mocked = flexmock()
        flexmock(mocked).should_receive("method1").and_return(2).and_return(3).one_by_one()
        assert mocked.method1() == 2
        assert mocked.method1() == 3
        assert mocked.method1() == 2
        assert mocked.method1() == 3

    def test_one_by_one_called_before_and_return_multiple_values(self):
        mocked = flexmock()
        mocked.should_receive("method1").one_by_one().and_return(3, 4)
        assert mocked.method1() == 3
        assert mocked.method1() == 4
        assert mocked.method1() == 3
        assert mocked.method1() == 4

    def test_one_by_one_called_before_and_return_one_value(self):
        mocked = flexmock()
        mocked.should_receive("method1").one_by_one().and_return(4).and_return(5)
        assert mocked.method1() == 4
        assert mocked.method1() == 5
        assert mocked.method1() == 4
        assert mocked.method1() == 5

    def test_flexmock_should_mix_multiple_return_values_with_exceptions(self):
        class Foo:
            def method1(self):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method1").and_return(1).and_raise(Exception)
        assert instance.method1() == 1
        with assert_raises(Exception, ""):
            instance.method1()
        assert instance.method1() == 1
        with assert_raises(Exception, ""):
            instance.method1()

    def test_flexmock_should_mock_new_instances_with_multiple_params(self):
        class User:
            pass

        class Group:
            def __init__(self, arg, arg2):
                pass

        user = User()
        flexmock(Group).new_instances(user)
        assert user is Group(1, 2)

    def test_flexmock_function_should_return_previously_mocked_object(self):
        class User:
            pass

        user = User()
        instance = flexmock(user)
        assert instance == user
        assert instance == flexmock(user)

    def test_flexmock_should_not_return_class_object_if_mocking_instance(self):
        class User:
            def method(self):
                pass

        user = User()
        user2 = User()
        class_mock = flexmock(User).should_receive("method").and_return("class").mock()
        user_mock = flexmock(user).should_receive("method").and_return("instance").mock()
        assert class_mock is not user_mock
        assert user.method() == "instance"
        assert user2.method() == "class"



"""Bogus module for testing."""
# pylint: disable=missing-docstring,disallowed-name,invalid-name,no-self-use

MODULE_LEVEL_ATTRIBUTE = "test"


class SomeClass:
    CLASS_VALUE = "class_method"

    def __init__(self):
        self.instance_value = "instance_method"

    @classmethod
    def class_method(cls):
        return cls.CLASS_VALUE

    @classmethod
    def class_method_with_args(cls, a):
        return a

    @staticmethod
    def static_method():
        return "static_method"

    @staticmethod
    def static_method_with_args(a):
        return a

    def instance_method(self):
        return self.instance_value

    def instance_method_with_args(self, a):
        return a


class DerivedClass(SomeClass):
    pass


class ModuleClass:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def a(self):
        return self.x + self.y


def module_function(x, y):
    return x - y


def kwargs_only_func1(foo, *, bar, baz=5):
    return foo + bar + baz


def kwargs_only_func2(foo, *, bar, baz):
    return foo + bar + baz



"""Flexmock public API."""
# pylint: disable=no-self-use,too-many-lines
import inspect
import re
import sys
import types
from types import BuiltinMethodType, TracebackType
from typing import Any, Callable, Dict, Iterator, List, NoReturn, Optional, Tuple, Type

from flexmock.exceptions import (
    CallOrderError,
    ExceptionClassError,
    ExceptionMessageError,
    FlexmockError,
    MethodCallError,
    MethodSignatureError,
    MockBuiltinError,
    StateError,
)

AT_LEAST = "at least"
AT_MOST = "at most"
EXACTLY = "exactly"
SPECIAL_METHODS = (classmethod, staticmethod)
UPDATED_ATTRS = ["should_receive", "should_call", "new_instances"]
DEFAULT_CLASS_ATTRIBUTES = [attr for attr in dir(type) if attr not in dir(type("", (object,), {}))]
# Fix Python 3.6 does not have re.Pattern type
RE_TYPE = type(re.compile(""))


class ReturnValue:
    """ReturnValue"""

    def __init__(
        self, value: Optional[Any] = None, raises: Optional[Type[BaseException]] = None
    ) -> None:
        self.value = value
        self.raises = raises

    def __str__(self) -> str:
        if self.raises:
            return f"{self.raises}({_arg_to_str(self.value)})"
        if not isinstance(self.value, tuple):
            return str(_arg_to_str(self.value))
        if len(self.value) == 1:
            return str(_arg_to_str(self.value[0]))
        values = ", ".join([_arg_to_str(x) for x in self.value])
        return f"({values})"


class Mock:
    """Fake object class returned by the `flexmock()` function."""

    def __init__(self, **kwargs: Any) -> None:
        """Mock constructor.

        Args:
            **kwargs: dict of attribute/value pairs used to initialize the mock
                object.
        """
        self._object: Any = self
        for attr, value in kwargs.items():
            if isinstance(value, property):
                setattr(self.__class__, attr, value)
            else:
                setattr(self, attr, value)

    def __enter__(self) -> Any:
        return self._object

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        pass

    def __call__(self, *args: Any, **kwargs: Any) -> "Mock":
        """Make mocks callable with and without parentheses.

        If `Mock` is not callable, it is difficult to mock attributes that
        should work with and without parentheses.
        """
        return self

    def __iter__(self) -> Iterator[Any]:
        """Makes the mock object iterable.

        Call the instance's version of __iter__ if available, otherwise yield self.
        """
        if (
            hasattr(self, "__dict__")
            and isinstance(self.__dict__, dict)
            and "__iter__" in self.__dict__
        ):
            for item in self.__dict__["__iter__"](self):
                yield item
        else:
            yield self

    def should_receive(self, name: str) -> "Expectation":
        """Replaces the specified attribute with a fake.

        Args:
            name: Name of the attribute to replace.

        Returns:
            Expectation object which can be used to modify the expectations
                on the fake attribute.

        Examples:
            >>> flexmock(plane).should_receive("fly").and_return("vooosh!").once()
            <flexmock._api.Expectation object at ...>
            >>> plane.fly()
            'vooosh!'
        """
        if name in UPDATED_ATTRS:
            raise FlexmockError("unable to replace flexmock methods")

        chained_methods = None
        if "." in name:
            name, chained_methods = name.split(".", 1)
        name = self._update_name_if_mangled(name)
        self._ensure_object_has_named_attribute(name)

        if chained_methods:
            if not isinstance(self._object, Mock) and not hasattr(
                getattr(self._object, name), "__call__"
            ):
                # Create a partial mock if the given name is callable
                # this allows chaining attributes
                return_value = _create_partial_mock(getattr(self._object, name))
            else:
                return_value = Mock()
            self._create_expectation(name, return_value)
            return return_value.should_receive(chained_methods)

        return self._create_expectation(name)

    def _update_name_if_mangled(self, name: str) -> str:
        """This allows flexmock to mock methods with name mangling."""
        if name.startswith("__") and not name.endswith("__") and not inspect.ismodule(self._object):
            class_name: str
            if inspect.isclass(self._object):
                class_name = self._object.__name__
            else:
                class_name = self._object.__class__.__name__
            name = f"_{class_name.lstrip('_')}__{name.lstrip('_')}"
        return name

    def _ensure_object_has_named_attribute(self, name: str) -> None:
        if not isinstance(self._object, Mock) and not self._hasattr(self._object, name):
            if hasattr(self._object, "__name__"):
                obj_name = self._object.__name__
            else:
                obj_name = str(self._object)
            raise FlexmockError(f"{obj_name} does not have attribute '{name}'")

    def _hasattr(self, obj: Any, name: str) -> bool:
        """Ensure hasattr checks don't create side-effects for properties."""
        if not inspect.isclass(obj) and hasattr(obj, "__dict__") and name not in obj.__dict__:
            if name in DEFAULT_CLASS_ATTRIBUTES:
                return False  # avoid false positives for things like __call__
            return hasattr(obj.__class__, name)
        return hasattr(obj, name)

    def should_call(self, name: str) -> "Expectation":
        """Creates a spy.

        This means that the original method will be called rather than the fake
        version. However, we can still keep track of how many times it is called
        and with what arguments, and apply expectations accordingly.

        `should_call` is meaningless/not allowed for non-callable attributes.

        Args:
            name: Name of the method to spy.

        Returns:
            Expectation object.

        Examples:
            >>> flexmock(plane).should_call("repair").once()
            <flexmock._api.Expectation object at ...>
            >>> plane.repair("wing")
        """
        if isinstance(self._object, Mock) and not hasattr(self._object, name):
            raise FlexmockError(
                f"Mock object does not have attribute '{name}'. "
                f'Did you mean to call should_receive("{name}") instead?'
            )
        expectation = self.should_receive(name)
        return expectation.replace_with(expectation.__dict__["_original"])

    def new_instances(self, *args: Any) -> "Expectation":
        """Overrides `__new__` method on a class to return custom objects.

        Alias for `should_receive('__new__').and_return(args).one_by_one()`.

        Args:
            *args: Objects to return on each successive call to `__new__`.

        Returns:
            Expectation object.

        Examples:

            >>> fake_plane = flexmock(model="fake")
            >>> flexmock(Plane).new_instances(fake_plane)
            <flexmock._api.Expectation object at ...>
            >>> Plane().model
            'fake'

            It is also possible to return different fake objects in a sequence:

            >>> fake_plane1 = flexmock(model="fake1")
            >>> fake_plane2 = flexmock(model="fake2")
            >>> flexmock(Plane).new_instances(fake_plane1, fake_plane2)
            <flexmock._api.Expectation object at ...>
            >>> Plane().model
            'fake1'
            >>> Plane().model
            'fake2'
        """
        if inspect.isclass(self._object):
            return self.should_receive("__new__").and_return(args).one_by_one()
        raise FlexmockError("new_instances can only be called on a class mock")

    def _create_expectation(self, name: str, return_value: Optional[Any] = None) -> "Expectation":
        expectation = self._get_or_create_expectation(name, return_value)
        FlexmockContainer.add_expectation(self, expectation)

        if _isproperty(self._object, name):
            self._update_property(expectation, name)
        elif (
            isinstance(self._object, Mock)
            or hasattr(getattr(self._object, name), "__call__")
            or inspect.isclass(getattr(self._object, name))
        ):
            self._update_method(expectation, name)
        else:
            self._update_attribute(expectation, name, return_value)
        return expectation

    def _get_or_create_expectation(
        self, name: str, return_value: Optional[Any] = None
    ) -> "Expectation":
        saved_expectations = FlexmockContainer.get_expectations_with_name(self, name)
        if saved_expectations:
            # If there is already an expectation for the same name, get the
            # original object from the FIRST saved expectation.
            return Expectation(
                self._object,
                name=name,
                return_value=return_value,
                original=saved_expectations[0].__dict__.get("_original"),
                method_type=saved_expectations[0].__dict__.get("_method_type"),
            )
        return Expectation(self._object, name=name, return_value=return_value)

    def _create_placeholder_mock_for_proper_teardown(
        self, obj: Any, name: str, original: Any
    ) -> None:
        """Ensures that the given function is replaced on teardown."""
        mock = Mock()
        mock._object = obj
        expectation = Expectation(obj, name=name, original=original)
        FlexmockContainer.add_expectation(mock, expectation)

    def _update_method(self, expectation: "Expectation", name: str) -> None:
        method_instance = self._create_mock_method(name)
        if self._hasattr(self._object, name) and not hasattr(expectation, "_original"):
            expectation._update_original(name, self._object)
            expectation._method_type = self._get_method_type(name, expectation._original)
            if expectation._method_type in SPECIAL_METHODS:
                expectation._original_function = getattr(self._object, name)
        if not inspect.isclass(self._object) or expectation._method_type in SPECIAL_METHODS:
            method_instance = types.MethodType(method_instance, self._object)
        expectation._local_override = _setattr(self._object, name, method_instance)
        if (
            expectation._local_override
            and not inspect.isclass(self._object)
            and not isinstance(self._object, Mock)
            and hasattr(self._object.__class__, name)
        ):
            self._update_class_for_magic_builtins(name)

    def _get_method_type(self, name: str, method: Callable[..., Any]) -> Any:
        """Get method type of the original method.

        Method type is saved because after mocking the base class, it is difficult to determine
        the original method type.
        """
        method_type = self._get_saved_method_type(name, method)
        if method_type is not None:
            return method_type
        if _is_class_method(method, name):
            method_type = classmethod
        elif _is_static_method(self._object, name):
            method_type = staticmethod
        else:
            method_type = type(method)
        setattr(self._object, f"{name}__flexmock__method_type", method_type)
        return method_type

    def _get_saved_method_type(self, name: str, method: Callable[..., Any]) -> Optional[Any]:
        """Check method type of the original method if it was saved to the class or base class."""
        bound_to = getattr(method, "__self__", None)
        if bound_to is not None and inspect.isclass(bound_to):
            # Check if the method type was saved in a base class
            for cls in inspect.getmro(bound_to):
                method_type = vars(cls).get(f"{name}__flexmock__method_type")
                if method_type:
                    return method_type
        return None

    def _update_class_for_magic_builtins(self, name: str) -> None:
        """Fixes method resolution order for built-in methods.

        Replacing magic builtins on instances has no effect as the one attached
        to the class takes precedence. To work around it, we update the class'
        method to check if the instance in question has one in its own __dict__
        and call that instead.
        """
        if not (name.startswith("__") and name.endswith("__") and len(name) > 4):
            return
        original = getattr(self._object.__class__, name)

        def updated(self: Any, *kargs: Any, **kwargs: Any) -> Any:
            if (
                hasattr(self, "__dict__")
                and isinstance(self.__dict__, dict)
                and name in self.__dict__
            ):
                return self.__dict__[name](*kargs, **kwargs)
            return original(self, *kargs, **kwargs)

        setattr(self._object.__class__, name, updated)
        if updated.__code__ != original.__code__:
            self._create_placeholder_mock_for_proper_teardown(
                self._object.__class__, name, original
            )

    def _update_attribute(
        self, expectation: "Expectation", name: str, return_value: Optional[Any] = None
    ) -> None:
        expectation._callable = False
        if self._hasattr(self._object, name) and not hasattr(expectation, "_original"):
            expectation._update_original(name, self._object)
        expectation._local_override = _setattr(self._object, name, return_value)

    def _update_property(self, expectation: "Expectation", name: str) -> None:
        new_name = f"_flexmock__{name}"
        obj = self._object
        if not inspect.isclass(obj):
            obj = obj.__class__

        expectation._callable = False
        original = getattr(obj, name)

        @property  # type: ignore
        def updated(self: Any) -> Any:
            if (
                hasattr(self, "__dict__")
                and isinstance(self.__dict__, dict)
                and name in self.__dict__
            ):
                return self.__dict__[name]
            # Return original for instances that are not mocked
            return getattr(self, new_name)

        setattr(obj, name, updated)
        if not hasattr(obj, new_name):
            # don't try to double update
            FlexmockContainer.add_teardown_property(obj, new_name)
            setattr(obj, new_name, original)
            self._create_placeholder_mock_for_proper_teardown(obj, name, original)

    def _create_mock_method(self, name: str) -> Callable[..., Any]:
        def _handle_exception_matching(expectation: Expectation) -> None:
            # pylint: disable=misplaced-bare-raise
            return_values = expectation._return_values
            if return_values:
                raised, instance = sys.exc_info()[:2]
                assert raised, "no exception was raised"
                message = str(instance)
                expected = return_values[0].raises
                if not expected:
                    raise
                args = return_values[0].value
                if inspect.isclass(expected):
                    assert isinstance(args, dict)
                    expected_instance = expected(*args["kargs"], **args["kwargs"])
                    expected_message = str(expected_instance)
                    if expected is not raised and expected not in raised.__bases__:
                        raise ExceptionClassError(
                            f"Raised exception for call {expectation._name} "
                            "did not match expectation:\n"
                            f"  Expected:\t{expected}\n"
                            f"  Raised:\t{raised}"
                        )
                    if args["kargs"] and isinstance(args["kargs"][0], RE_TYPE):
                        if not args["kargs"][0].search(message):
                            raise ExceptionMessageError(
                                f"Error message mismatch with raised {expected.__name__}:\n"
                                f"  Expected pattern:\n\t/{args['kargs'][0].pattern}/\n"
                                f"  Received message:\n\t'{message}'"
                            )
                    elif expected_message and expected_message != message:
                        raise (
                            ExceptionMessageError(
                                f"Error message mismatch with raised {expected.__name__}:\n"
                                f"  Expected message:\n\t'{message}'\n"
                                f"  Received message:\n\t'{expected_message}'"
                            )
                        )
                elif expected is not raised:
                    raise ExceptionClassError(
                        f"Raised exception for call {expectation._name} "
                        f"did not match expectation:\n"
                        f"  Expected:\t{repr(expected)}\n"
                        f"  Raised:\t{raised}\n\n"
                        "Did you try to call and_raise with an instance?\n"
                        'Instead of and_raise(Exception("arg")), try and_raise(Exception, "arg")'
                    )
            else:
                raise

        def match_return_values(expected: Any, received: Any) -> bool:
            if not isinstance(expected, tuple):
                expected = (expected,)
            if not isinstance(received, tuple):
                received = (received,)
            if len(received) != len(expected):
                return False
            for i, val in enumerate(received):
                if not _arguments_match(val, expected[i]):
                    return False
            return True

        def pass_thru(
            expectation: Expectation, runtime_self: Any, *kargs: Any, **kwargs: Any
        ) -> Any:
            return_values = None
            try:
                original = expectation._original
                _mock = expectation._mock
                if inspect.isclass(_mock):
                    if expectation._method_type in SPECIAL_METHODS:
                        original = expectation._original_function
                        return_values = original(*kargs, **kwargs)
                    else:
                        return_values = original(runtime_self, *kargs, **kwargs)
                else:
                    return_values = original(*kargs, **kwargs)
            except Exception:
                return _handle_exception_matching(expectation)
            expected_values = expectation._return_values
            if expected_values and not match_return_values(expected_values[0].value, return_values):
                expected_value = expected_values[0].value
                # Display strings with quotes in the error message
                if isinstance(return_values, str):
                    return_values = repr(return_values)
                if isinstance(expected_value, str):
                    expected_value = repr(expected_value)
                raise (
                    MethodSignatureError(
                        f"Returned values for call {expectation._name} did not match expectation:\n"
                        f"  Expected:\t{expected_value}\n"
                        f"  Returned:\t{return_values}"
                    )
                )
            return return_values

        def _handle_matched_expectation(
            expectation: Expectation, runtime_self: Any, *kargs: Any, **kwargs: Any
        ) -> Any:
            if not expectation._runnable():
                raise StateError(
                    f"{name} expected to be called when {expectation._get_runnable()} is True"
                )
            expectation._times_called += 1
            expectation._verify(final=False)
            _pass_thru = expectation._pass_thru
            _replace_with = expectation._replace_with
            if _pass_thru:
                return pass_thru(expectation, runtime_self, *kargs, **kwargs)
            if _replace_with:
                return _replace_with(*kargs, **kwargs)
            return_values = expectation._return_values
            if return_values:
                return_value = return_values[0]
                del return_values[0]
                return_values.append(return_value)
            else:
                return_value = ReturnValue()
            if return_value.raises:
                if inspect.isclass(return_value.raises):
                    args = return_value.value
                    assert isinstance(args, dict)
                    raise return_value.raises(*args["kargs"], **args["kwargs"])
                raise return_value.raises  # pylint: disable=raising-bad-type
            return return_value.value

        def mock_method(runtime_self: Any, *kargs: Any, **kwargs: Any) -> Any:
            arguments = {"kargs": kargs, "kwargs": kwargs}
            expectation = FlexmockContainer.get_flexmock_expectation(self, name, arguments)
            if expectation:
                return _handle_matched_expectation(expectation, runtime_self, *kargs, **kwargs)
            # inform the user which expectation(s) for the method were _not_ matched
            saved_expectations = reversed(FlexmockContainer.get_expectations_with_name(self, name))
            error_msg = (
                f"Arguments for call {name} did not match expectations:\n"
                f"  Received call:\t{_format_args(name, arguments)}\n"
            )
            if saved_expectations:
                error_msg += "\n".join(
                    f"  Expected call[{index}]:\t{_format_args(name, expectation._args)}"
                    for index, expectation in enumerate(saved_expectations, 1)
                )
            raise MethodSignatureError(error_msg)

        return mock_method


def flexmock_teardown() -> None:
    """Performs flexmock-specific teardown tasks."""
    saved = {}
    instances = []
    classes = []
    for mock_object, expectations in FlexmockContainer.flexmock_objects.items():
        saved[mock_object] = expectations[:]
        for expectation in expectations:
            expectation._reset()
        for expectation in expectations:
            # Remove method type attributes set by flexmock. This needs to be done after
            # resetting all the expectations because method type is needed in expectation teardown.
            if inspect.isclass(mock_object) or hasattr(mock_object, "__class__"):
                try:
                    delattr(mock_object._object, f"{expectation._name}__flexmock__method_type")
                except (AttributeError, TypeError):
                    pass
    for mock in saved:
        obj = mock._object
        if not isinstance(obj, Mock) and not inspect.isclass(obj):
            instances.append(obj)
        if inspect.isclass(obj):
            classes.append(obj)
    for obj in instances + classes:
        for attr in UPDATED_ATTRS:
            try:
                obj_dict = obj.__dict__
                if obj_dict[attr].__code__ is Mock.__dict__[attr].__code__:
                    del obj_dict[attr]
            except Exception:
                try:
                    if getattr(obj, attr).__code__ is Mock.__dict__[attr].__code__:
                        delattr(obj, attr)
                except AttributeError:
                    pass
    FlexmockContainer.teardown_properties()
    FlexmockContainer.reset()

    # make sure this is done last to keep exceptions here from breaking
    # any of the previous steps that cleanup all the changes
    for mock_object, expectations in saved.items():
        for expectation in expectations:
            expectation._verify()


class Expectation:
    """Holds expectations about methods.

    The information contained in the Expectation object includes method name,
    its argument list, return values, and any exceptions that the method might
    raise.
    """

    def __init__(
        self,
        mock: Mock,
        name: Optional[str] = None,
        return_value: Optional[Any] = None,
        original: Optional[Any] = None,
        method_type: Optional[Any] = None,
    ) -> None:
        if original is not None:
            self._original = original

        self._name = name
        self._times_called: int = 0
        self._modifier: str = EXACTLY
        self._args: Optional[Dict[str, Any]] = None
        self._method_type = method_type
        self._argspec: Optional[inspect.FullArgSpec] = None
        self._return_values = [ReturnValue(return_value)] if return_value is not None else []
        self._replace_with: Optional[Callable[..., Any]] = None
        self._original_function: Optional[Callable[..., Any]] = None
        self._expected_calls: Dict[str, Optional[int]] = {
            EXACTLY: None,
            AT_LEAST: None,
            AT_MOST: None,
        }
        self._runnable: Callable[..., bool] = lambda: True
        self._mock = mock
        self._pass_thru = False
        self._ordered = False
        self._one_by_one = False
        self._verified = False
        self._callable = True
        self._local_override = False
        # Remove in 1.0. Issue #96
        self._called_deprecated_property = False

    def __str__(self) -> str:
        args = _format_args(str(self._name), self._args)
        return_values = ", ".join(str(x) for x in self._return_values)
        return f"{args} -> ({return_values})"

    def __getattribute__(self, name: str) -> Any:
        # Workaround to raise an error if once, twice, or never are called
        # without parenthesis.
        if name in ("once", "twice", "never", "mock"):
            self._called_deprecated_property = True
        return object.__getattribute__(self, name)

    def _get_runnable(self) -> str:
        """Ugly hack to get the name of when() condition from the source code."""
        name = "condition"
        try:
            source = inspect.getsource(self._runnable)
            if "when(" in source:
                name = source.split("when(")[1].split(")")[0]
            elif "def " in source:
                name = source.split("def ")[1].split("(")[0]
        except Exception:
            # couldn't get the source, oh well
            pass
        return name

    def _verify_signature_match(self, *kargs: Any, **kwargs: Any) -> None:
        if isinstance(self._mock, Mock):
            return  # no sense in enforcing this for fake objects
        allowed = self._argspec
        args_len = len(allowed.args)

        # self is the first expected argument
        has_self = allowed.args and allowed.args[0] == "self"
        # Builtin methods take `self` as the first argument but `inspect.ismethod` returns False
        # so we need to check for them explicitly
        is_builtin_method = isinstance(self._original, BuiltinMethodType) and has_self
        # Methods take `self` if not a staticmethod
        is_method = inspect.ismethod(self._original) and self._method_type is not staticmethod
        # Class init takes `self`
        is_class = inspect.isclass(self._original)
        # When calling class methods or instance methods on a class method takes `cls`
        is_class_method = (
            inspect.isfunction(self._original)
            and inspect.isclass(self._mock)
            and self._method_type is not staticmethod
        )
        if is_builtin_method or is_method or is_class or is_class_method:
            # Do not count `self` or `cls`.
            args_len -= 1

        minimum = args_len - (allowed.defaults and len(allowed.defaults) or 0)
        maximum = None
        if allowed.varargs is None and allowed.varkw is None:
            maximum = args_len
        total_positional = len(kargs + tuple(a for a in kwargs if a in allowed.args))
        named_optionals = [
            a
            for a in kwargs
            if allowed.defaults
            if a in allowed.args[len(allowed.args) - len(allowed.defaults) :]
        ]
        if allowed.defaults and total_positional == minimum and named_optionals:
            minimum += len(named_optionals)
        if total_positional < minimum:
            arguments = "argument" if minimum == 1 else "arguments"
            raise MethodSignatureError(
                f"{self._name} requires at least {minimum} {arguments}, "
                f"expectation provided {total_positional}"
            )
        if maximum is not None and total_positional > maximum:
            arguments = "argument" if maximum == 1 else "arguments"
            raise MethodSignatureError(
                f"{self._name} requires at most {maximum} {arguments}, "
                f"expectation provided {total_positional}"
            )
        if args_len == len(kargs) and any(a for a in kwargs if a in allowed.args):
            given_args = [a for a in kwargs if a in allowed.args]
            arguments = "argument" if len(given_args) == 1 else "arguments"
            raise MethodSignatureError(
                f"{given_args} already given as positional {arguments} to {self._name}"
            )
        if not allowed.varkw and any(
            a for a in kwargs if a not in allowed.args + allowed.kwonlyargs
        ):
            invalid_arg = [a for a in kwargs if a not in allowed.args + allowed.kwonlyargs][0]
            raise MethodSignatureError(
                f"{invalid_arg} is not a valid keyword argument to {self._name}"
            )
        # check that kwonlyargs that don't have default value specified are provided
        required_kwonlyargs = [
            a for a in allowed.kwonlyargs if a not in (allowed.kwonlydefaults or {})
        ]
        missing_kwonlyargs = [a for a in required_kwonlyargs if a not in kwargs]
        if missing_kwonlyargs:
            arguments = "argument" if len(missing_kwonlyargs) == 1 else "arguments"
            missing_args = '", "'.join(missing_kwonlyargs)
            raise MethodSignatureError(
                f'{self._name} requires keyword-only {arguments} "{missing_args}"'
            )

    def _update_original(self, name: str, obj: Any) -> None:
        if hasattr(obj, "__dict__") and name in obj.__dict__:
            self._original = obj.__dict__[name]
        else:
            self._original = getattr(obj, name)
        self._update_argspec()

    def _update_argspec(self) -> None:
        original = self.__dict__.get("_original")
        if original:
            try:
                self._argspec = inspect.getfullargspec(original)
            except TypeError:
                # built-in function: fall back to stupid processing and hope the
                # builtins don't change signature
                pass

    def _normalize_named_args(self, *kargs: Any, **kwargs: Any) -> Dict[str, Any]:
        argspec = self._argspec
        default = {"kargs": kargs, "kwargs": kwargs}
        if not argspec:
            return default
        ret: Dict[str, Any] = {"kargs": (), "kwargs": kwargs}
        if inspect.ismethod(self._original):
            args = argspec.args[1:]
        else:
            args = argspec.args
        for i, arg in enumerate(kargs):
            if len(args) <= i:
                return default
            ret["kwargs"][args[i]] = arg
        return ret

    def __raise(self, exception: Type[BaseException], message: str) -> NoReturn:
        """Safe internal raise implementation.

        In case we're patching builtins, it's important to reset the
        expectation before raising any exceptions or else things like
        open() might be stubbed out and the resulting runner errors are very
        difficult to diagnose.
        """
        self._reset()
        raise exception(message)

    def _match_args(self, given_args: Any) -> bool:
        """Check if the set of given arguments matches this expectation."""
        expected_args = self._args
        given_args = self._normalize_named_args(*given_args["kargs"], **given_args["kwargs"])
        if expected_args == given_args or expected_args is None:
            return True
        if (
            len(given_args["kargs"]) != len(expected_args["kargs"])
            or len(given_args["kwargs"]) != len(expected_args["kwargs"])
            or (sorted(given_args["kwargs"].keys()) != sorted(expected_args["kwargs"].keys()))
        ):
            return False
        for i, arg in enumerate(given_args["kargs"]):
            if not _arguments_match(arg, expected_args["kargs"][i]):
                return False
        for key, value in given_args["kwargs"].items():
            if not _arguments_match(value, expected_args["kwargs"][key]):
                return False
        return True

    def mock(self) -> Mock:
        """Return the mock associated with this expectation.

        Returns:
            Mock associated with this expectation.

        Examples:
            >>> plane = flexmock(plane).should_receive("fly").mock()
        """
        self._called_deprecated_property = False
        return self._mock

    def with_args(self, *args: Any, **kwargs: Any) -> "Expectation":
        """Override the arguments used to match this expectation's method.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            Match calls with no arguments:

            >>> flexmock(plane).should_receive("fly").with_args()
            <flexmock._api.Expectation object at ...>

            Match a single argument:

            >>> flexmock(plane).should_receive("fly").with_args("east")
            <flexmock._api.Expectation object at ...>

            Match keyword arguments:

            >>> flexmock(plane).should_receive("fly").with_args("up", destination="Oslo")
            <flexmock._api.Expectation object at ...>

            Match argument type:

            >>> flexmock(plane).should_receive("fly").with_args(str)
            <flexmock._api.Expectation object at ...>

            Match a string using a compiled regular expression:

            >>> regex = re.compile("^(up|down)$")
            >>> flexmock(plane).should_receive("fly").with_args(regex)
            <flexmock._api.Expectation object at ...>
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use with_args() with attribute stubs")
        self._update_argspec()
        if self._argspec:
            # do this outside try block as TypeError is way too general and catches
            # unrelated errors in the verify signature code
            self._verify_signature_match(*args, **kwargs)
            self._args = self._normalize_named_args(*args, **kwargs)
        else:
            self._args = {"kargs": args, "kwargs": kwargs}
        return self

    def and_return(self, *values: Any) -> "Expectation":
        """Override the return value of this expectation's method.

        When `and_return` is given multiple times, each value provided is returned
        on successive invocations of the method. It is also possible to mix
        `and_return` with `and_raise` in the same manner to alternate between returning
        a value and raising and exception on different method invocations.

        When combined with the `one_by_one` modifier, value is treated as a list of
        values to be returned in the order specified by successive calls to this
        method rather than a single list to be returned each time.

        Args:
            *values: Optional list of return values, defaults to `None` if not given.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            Provide return values for mocks:

            >>> flexmock(plane).should_receive("land").and_return("landed!").once()
            <flexmock._api.Expectation object at ...>
            >>> plane.land()
            'landed!'

            Match specific return values with spies:

            >>> flexmock(plane).should_call("passenger_count").and_return(3)
            <flexmock._api.Expectation object at ...>
            >>> plane.passenger_count()
            3
        """
        if not values:
            value = None
        elif len(values) == 1:
            value = values[0]
        else:
            value = values

        if not self._callable:
            _setattr(self._mock, str(self._name), value)
            return self

        return_values = self._return_values
        if not self._one_by_one:
            value = ReturnValue(value)
            return_values.append(value)
        else:
            try:
                return_values.extend([ReturnValue(v) for v in value])  # type: ignore
            except TypeError:
                return_values.append(ReturnValue(value))
        return self

    def times(self, number: int) -> "Expectation":
        """Number of times this expectation's method is expected to be called.

        There are also 3 aliases for the `times()` method:

          - `once()` -> `times(1)`
          - `twice()` -> `times(2)`
          - `never()` -> `times(0)`

        Args:
            number: Expected call count.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("fly").times(1)
            <flexmock._api.Expectation object at ...>
            >>> plane.fly()

            >>> flexmock(plane).should_call("land").times(2)
            <flexmock._api.Expectation object at ...>
            >>> plane.land()
            >>> plane.land()
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use times() with attribute stubs")
        expected_calls = self._expected_calls
        modifier = self._modifier
        expected_calls[modifier] = number
        return self

    def once(self) -> "Expectation":
        """Expect expectation's method to be called once.

        Alias for `times(1)`.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("land").once()
            <flexmock._api.Expectation object at ...>
            >>> plane.land()
        """
        self._called_deprecated_property = False
        return self.times(1)

    def twice(self) -> "Expectation":
        """Expect expectation's method to be called twice.

        Alias for `times(2)`.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("fly").twice()
            <flexmock._api.Expectation object at ...>
            >>> plane.fly()
            >>> plane.fly()
        """
        self._called_deprecated_property = False
        return self.times(2)

    def never(self) -> "Expectation":
        """Expect expectation's method to never be called.

        Alias for `times(0)`.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("crash").never()
            <flexmock._api.Expectation object at ...>
            >>> plane.crash()
            Traceback (most recent call last):
              ...
            flexmock.exceptions.MethodCallError: crash() expected to be called...
        """
        self._called_deprecated_property = False
        return self.times(0)

    def one_by_one(self) -> "Expectation":
        """Modifies the return value to be treated as a list of return values.

        Each value in the list is returned on successive invocations of the method.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("pilot").and_return("pilot1", "pilot2").one_by_one()
            <flexmock._api.Expectation object at ...>
            >>> plane.pilot()
            'pilot1'
            >>> plane.pilot()
            'pilot2'
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use one_by_one() with attribute stubs")
        if not self._one_by_one:
            self._one_by_one = True
            return_values = self._return_values
            saved_values = return_values[:]
            self._return_values = return_values = []
            for value in saved_values:
                try:
                    for val in value.value:  # type: ignore
                        return_values.append(ReturnValue(val))
                except TypeError:
                    return_values.append(value)
        return self

    def at_least(self) -> "Expectation":
        """Modifies the associated `times()` expectation.

        When given, an exception will only be raised if the method is called less
        than `times()` specified. Does nothing if `times()` is not given.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("fly").at_least().once()
            <flexmock._api.Expectation object at ...>
            >>> plane.fly("east")
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use at_least() with attribute stubs")
        expected_calls = self._expected_calls
        modifier = self._modifier
        if expected_calls[AT_LEAST] is not None or modifier == AT_LEAST:
            self.__raise(FlexmockError, "cannot use at_least modifier twice")
        if modifier == AT_MOST and expected_calls[AT_MOST] is None:
            self.__raise(FlexmockError, "cannot use at_least with at_most unset")
        self._modifier = AT_LEAST
        return self

    def at_most(self) -> "Expectation":
        """Modifies the associated `times()` expectation.

        When given, an exception will only be raised if the method is called more
        than `times()` specified. Does nothing if `times()` is not given.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("land").at_most().once()
            <flexmock._api.Expectation object at ...>
            >>> plane.land()
            >>> plane.land()
            Traceback (most recent call last):
              ...
            flexmock.exceptions.MethodCallError: land() expected to be called at most...
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use at_most() with attribute stubs")
        expected_calls = self._expected_calls
        modifier = self._modifier
        if expected_calls[AT_MOST] is not None or modifier == AT_MOST:
            self.__raise(FlexmockError, "cannot use at_most modifier twice")
        if modifier == AT_LEAST and expected_calls[AT_LEAST] is None:
            self.__raise(FlexmockError, "cannot use at_most with at_least unset")
        self._modifier = AT_MOST
        return self

    def ordered(self) -> "Expectation":
        """Makes the expectation respect the order of `should_receive` statements.

        An exception will be raised if methods are called out of order, determined
        by order of `should_receive` calls in the test.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("fly").with_args("east").ordered()
            <flexmock._api.Expectation object at ...>
            >>> flexmock(plane).should_receive("fly").with_args("west").ordered()
            <flexmock._api.Expectation object at ...>
            >>> plane.fly("west")
            Traceback (most recent call last):
              ...
            flexmock.exceptions.CallOrderError: fly("west") called before...
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use ordered() with attribute stubs")
        self._ordered = True
        FlexmockContainer.ordered.append(self)
        return self

    def when(self, func: Callable[..., Any]) -> "Expectation":
        """Sets an outside resource to be checked when asserting if a method
        should be called.

        Args:
            func: Function that indicates if the mocked or spied method should
                have been called.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("land").when(lambda: plane.is_flying)
            <flexmock._api.Expectation object at ...>
            >>> plane.is_flying
            True
            >>> plane.land()
            >>> plane.is_flying = False
            >>> plane.land()
            Traceback (most recent call last):
              ...
            flexmock.exceptions.StateError: land expected to be called when...
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use when() with attribute stubs")
        if not hasattr(func, "__call__"):
            self.__raise(FlexmockError, "when() parameter must be callable")
        self._runnable = func
        return self

    def and_raise(self, exception: Type[BaseException], *args: Any, **kwargs: Any) -> "Expectation":
        """Specifies the exception to be raised when this expectation is met.

        Args:
            exception: Exception class.
            *args: Positional arguments to pass to the exception.
            **kwargs: Keyword arguments to pass to the exception.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            Make a mocked method raise an exception instead of returning a value:

            >>> flexmock(plane).should_receive("fly").and_raise(BadWeatherException)
            <flexmock._api.Expectation object at ...>
            >>> plane.fly()
            Traceback (most recent call last):
              ...
            BadWeatherException

            Make a spy to verify that a specific exception is raised:

            >>> flexmock(plane).should_call("repair").and_raise(RuntimeError, "err msg")
            <flexmock._api.Expectation object at ...>
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use and_raise() with attribute stubs")
        if inspect.isclass(exception):
            try:
                exception(*args, **kwargs)
            except TypeError:
                self.__raise(
                    FlexmockError, f"can't initialize {exception} with the given arguments"
                )
        kargs = {"kargs": args, "kwargs": kwargs}
        return_values = self._return_values
        return_values.append(ReturnValue(raises=exception, value=kargs))
        return self

    def replace_with(self, function: Callable[..., Any]) -> "Expectation":
        """Gives a function to run instead of the mocked out one.

        Args:
            function: A callable that is used to replace the original function.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("set_speed").replace_with(lambda x: x == 5)
            <flexmock._api.Expectation object at ...>
            >>> plane.set_speed(5)
            True
            >>> plane.set_speed(4)
            False
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use replace_with() with attribute/property stubs")
        replace_with = self._replace_with
        original = self.__dict__.get("_original")
        if replace_with:
            self.__raise(FlexmockError, "replace_with cannot be specified twice")
        if function == original:
            self._pass_thru = True
        self._replace_with = function
        return self

    def and_yield(self, *args: Any) -> "Expectation":
        """Specifies the list of items to be yielded on successive method calls.

        In effect, the mocked object becomes a generator.

        Args:
            *args: Items to be yielded on successive method calls.

        Returns:
            Self, i.e. can be chained with other Expectation methods.

        Examples:
            >>> flexmock(plane).should_receive("flight_log").and_yield("fly", "land")
            <flexmock._api.Expectation object at ...>
            >>> log = plane.flight_log()
            >>> next(log)
            'fly'
            >>> next(log)
            'land'
        """
        if not self._callable:
            self.__raise(FlexmockError, "can't use and_yield() with attribute stubs")
        return self.and_return(iter(args))

    def _verify(self, final: bool = True) -> None:
        """Verify that this expectation has been met.

        Args:
            final: True if no further calls to this method expected
                (skip checking at_least expectations when False).

        Raises:
            MethodCallError Exception
        """
        failed, message = self._verify_number_of_calls(final)
        if failed and not self._verified:
            self._verified = True
            self.__raise(
                MethodCallError,
                (
                    f"{_format_args(str(self._name), self._args)} expected to be called "
                    f"{message}, called {self._times_called} "
                    f"{'time' if self._times_called == 1 else 'times'}"
                ),
            )

    def _verify_number_of_calls(self, final: bool) -> Tuple[bool, str]:
        if self._called_deprecated_property:
            raise FlexmockError(
                "Calling once, twice, never, or mock without parentheses has been deprecated"
            )
        failed = False
        message = ""
        called_exactly = self._expected_calls[EXACTLY]
        called_at_least = self._expected_calls[AT_LEAST]
        called_at_most = self._expected_calls[AT_MOST]
        times_called = self._times_called
        if called_exactly is not None:
            message = f"exactly {called_exactly}"
            if final:
                if times_called != called_exactly:
                    failed = True
            else:
                if times_called > called_exactly:
                    failed = True
            message += " time" if called_exactly == 1 else " times"
        else:
            if final and called_at_least is not None:
                message = f"at least {called_at_least}"
                if times_called < called_at_least:
                    failed = True
                message += " time" if called_at_least == 1 else " times"
            if called_at_most is not None:
                if message:
                    message += " and "
                message += f"at most {called_at_most}"
                if times_called > called_at_most:
                    failed = True
                message += " time" if called_at_most == 1 else " times"
        return failed, message

    def _reset(self) -> None:
        """Returns the methods overriden by this expectation to their originals."""
        _mock = self._mock
        if not isinstance(_mock, Mock):
            original = self.__dict__.get("_original")
            if original:
                # name may be unicode but pypy demands dict keys to be str
                name = str(self._name)
                if hasattr(_mock, "__dict__") and name in _mock.__dict__ and self._local_override:
                    delattr(_mock, name)
                elif (
                    hasattr(_mock, "__dict__")
                    and name in _mock.__dict__
                    and isinstance(_mock.__dict__, dict)
                ):
                    _mock.__dict__[name] = original
                else:
                    setattr(_mock, name, original)
        del self


class FlexmockContainer:
    """Holds global hash of object/expectation mappings."""

    flexmock_objects: Dict[Mock, List[Expectation]] = {}
    properties: Dict[Any, List[str]] = {}
    ordered: List[Expectation] = []
    last: Optional[Expectation] = None

    @classmethod
    def reset(cls) -> None:
        """Reset flexmock state."""
        cls.ordered = []
        cls.last = None
        cls.flexmock_objects = {}
        cls.properties = {}

    @classmethod
    def get_flexmock_expectation(
        cls, obj: Mock, name: str, args: Optional[Any] = None
    ) -> Optional[Expectation]:
        """Retrieves an existing matching expectation."""
        if args is None:
            args = {"kargs": (), "kwargs": {}}
        if not isinstance(args, dict):
            args = {"kargs": args, "kwargs": {}}
        if not isinstance(args["kargs"], tuple):
            args["kargs"] = (args["kargs"],)
        found = None
        for expectation in reversed(cls.flexmock_objects[obj]):
            if expectation._name == name and expectation._match_args(args):
                if expectation in cls.ordered or not expectation._ordered and not found:
                    found = expectation
        if found and found._ordered:
            cls._verify_call_order(found, args)
        return found

    @classmethod
    def _verify_call_order(cls, expectation: Expectation, args: Dict[str, Any]) -> None:
        next_method = cls.ordered.pop(0)
        cls.last = next_method
        if expectation is not next_method:
            raise CallOrderError(
                f"{_format_args(str(expectation._name), args)} called before "
                f"{_format_args(str(next_method._name), next_method._args)}"
            )

    @classmethod
    def add_expectation(cls, obj: Mock, expectation: Expectation) -> None:
        """Add expectation."""
        if obj in cls.flexmock_objects:
            cls.flexmock_objects[obj].append(expectation)
        else:
            cls.flexmock_objects[obj] = [expectation]

    @classmethod
    def get_expectations_with_name(cls, obj: Mock, name: str) -> List[Expectation]:
        """Get all expectations for given name."""
        return [x for x in FlexmockContainer.flexmock_objects.get(obj, []) if x._name == name]

    @classmethod
    def add_teardown_property(cls, obj: Any, name: str) -> None:
        """Add teardown property."""
        if obj in cls.properties:
            cls.properties[obj].append(name)
        else:
            cls.properties[obj] = [name]

    @classmethod
    def teardown_properties(cls) -> None:
        """Teardown properties."""
        for obj, names in cls.properties.items():
            for name in names:
                delattr(obj, name)


def flexmock(spec: Optional[Any] = None, **kwargs: Any) -> Mock:
    """Main entry point into the flexmock API.

    This function is used to either generate a new fake object or take
    an existing object (or class or module) and use it as a basis for
    a partial mock. In case of a partial mock, the passed in object
    is modified to support basic Mock class functionality making
    it unnecessary to make successive `flexmock()` calls on the same
    objects to generate new expectations.

    It is safe to call `flexmock()` on the same object again, flexmock will
    detect when an object has already been partially mocked and return it each
    time.

    Args:
        spec: Object, class, or module to mock.
        **kwargs: Method or return value pairs to attach to the object.

    Returns:
        Mock object if no spec is provided. Otherwise return the spec object.

    Examples:
        >>> mock = flexmock(Plane)
        >>> mock.should_receive("fly")
        <flexmock._api.Expectation object at ...>
    """
    if spec is not None:
        return _create_partial_mock(spec, **kwargs)
    # use this intermediate class to attach properties
    klass = type("MockClass", (Mock,), {})
    return klass(**kwargs)  # type: ignore


def _arg_to_str(arg: Any) -> str:
    if isinstance(arg, RE_TYPE):
        return f"/{arg.pattern}/"
    if isinstance(arg, str):
        return f'"{arg}"'
    return f"{arg}"


def _format_args(name: str, arguments: Optional[Dict[str, Any]]) -> str:
    if arguments is None:
        arguments = {"kargs": (), "kwargs": {}}
    kargs = ", ".join(_arg_to_str(arg) for arg in arguments["kargs"])
    kwargs = ", ".join(f"{k}={_arg_to_str(v)}" for k, v in arguments["kwargs"].items())
    if kargs and kwargs:
        args = f"{kargs}, {kwargs}"
    else:
        args = f"{kargs}{kwargs}"
    return f"{name}({args})"


def _create_partial_mock(obj_or_class: Any, **kwargs: Any) -> Mock:
    """Create partial mock."""
    matches = [x for x in FlexmockContainer.flexmock_objects if x._object is obj_or_class]
    if matches:
        mock = matches[0]
    else:
        mock = Mock()
        mock._object = obj_or_class
    for name, return_value in kwargs.items():
        if hasattr(return_value, "__call__"):
            mock.should_receive(name).replace_with(return_value)
        else:
            mock.should_receive(name).and_return(return_value)
    if not matches:
        FlexmockContainer.add_expectation(mock, Expectation(obj_or_class))
    if _attach_flexmock_methods(mock, Mock, obj_or_class) and not inspect.isclass(mock._object):
        mock = mock._object
    return mock


def _attach_flexmock_methods(mock: Mock, flexmock_class: Type[Mock], obj: Any) -> bool:
    try:
        for attr in UPDATED_ATTRS:
            if hasattr(obj, attr):
                if getattr(obj, attr).__code__ is not getattr(flexmock_class, attr).__code__:
                    return False
        for attr in UPDATED_ATTRS:
            _setattr(obj, attr, getattr(mock, attr))
    except TypeError as exc:
        raise MockBuiltinError(
            "Python does not allow you to mock builtin objects or modules. "
            "Consider wrapping it in a class you can mock instead"
        ) from exc
    except AttributeError as exc:
        raise MockBuiltinError(
            "Python does not allow you to mock instances of builtin objects. "
            "Consider wrapping it in a class you can mock instead"
        ) from exc
    return True


def _arguments_match(arg: Any, expected_arg: Any) -> bool:
    if expected_arg == arg:
        return True
    if inspect.isclass(expected_arg) and isinstance(arg, expected_arg):
        return True
    if isinstance(expected_arg, RE_TYPE) and expected_arg.search(arg):
        return True
    return False


def _setattr(obj: Any, name: str, value: Any) -> bool:
    """Ensure we use local __dict__ where possible."""
    local_override = False
    if hasattr(obj, "__dict__") and isinstance(obj.__dict__, dict):
        if name not in obj.__dict__:
            # Overriding attribute locally on an instance.
            local_override = True
        obj.__dict__[name] = value
    else:
        if inspect.isclass(obj) and not vars(obj).get(name):
            # Overriding derived attribute locally on a child class.
            local_override = True
        setattr(obj, name, value)
    return local_override


def _isproperty(obj: Any, name: str) -> bool:
    if isinstance(obj, Mock):
        return False
    if not inspect.isclass(obj) and hasattr(obj, "__dict__") and name not in obj.__dict__:
        attr = getattr(obj.__class__, name)
        if isinstance(attr, property):
            return True
    elif inspect.isclass(obj):
        attr = getattr(obj, name)
        if isinstance(attr, property):
            return True
    return False


def _is_class_method(method: Callable[..., Any], name: str) -> bool:
    """Check if a method is a classmethod.

    This function checks all the classes in the class method resolution in order
    to get the correct result for derived methods as well.
    """
    bound_to = getattr(method, "__self__", None)
    if not inspect.isclass(bound_to):
        return False
    for cls in inspect.getmro(bound_to):
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False


def _is_static_method(obj: Any, name: str) -> bool:
    try:
        return isinstance(inspect.getattr_static(obj, name), staticmethod)
    except AttributeError:
        # AttributeError is raised when mocking a proxied object
        if hasattr(obj, "__mro__"):
            for cls in inspect.getmro(obj):
                descriptor = vars(cls).get(name)
                if descriptor is not None:
                    return isinstance(descriptor, staticmethod)
        return False



"""Tests for call count modifiers."""
# pylint: disable=missing-docstring,no-self-use,no-member

from flexmock import exceptions, flexmock
from flexmock._api import flexmock_teardown
from tests.utils import assert_raises


class CallCountTestCase:
    def test_support_at_least_and_at_most_together(self):
        class Foo:
            def method(self):
                pass

        instance = Foo()
        flexmock(instance).should_call("method").at_least().once().at_most().twice()
        with assert_raises(
            exceptions.MethodCallError,
            "method() expected to be called at least 1 time and at most 2 times, called 0 times",
        ):
            flexmock_teardown()

        flexmock(instance).should_call("method").at_least().once().at_most().twice()
        instance.method()
        instance.method()
        with assert_raises(
            exceptions.MethodCallError,
            "method() expected to be called at most 2 times, called 3 times",
        ):
            instance.method()

        flexmock(instance).should_call("method").at_least().once().at_most().twice()
        instance.method()
        flexmock_teardown()

        flexmock(instance).should_call("method").at_least().once().at_most().twice()
        instance.method()
        instance.method()
        flexmock_teardown()

    def test_at_least_cannot_be_used_twice(self):
        class Foo:
            def method(self):
                pass

        expectation = flexmock(Foo).should_receive("method")
        with assert_raises(exceptions.FlexmockError, "cannot use at_least modifier twice"):
            expectation.at_least().at_least()

    def test_at_most_cannot_be_used_twice(self):
        class Foo:
            def method(self):
                pass

        expectation = flexmock(Foo).should_receive("method")
        with assert_raises(exceptions.FlexmockError, "cannot use at_most modifier twice"):
            expectation.at_most().at_most()

    def test_at_least_cannot_be_specified_until_at_most_is_set(self):
        class Foo:
            def method(self):
                pass

        expectation = flexmock(Foo).should_receive("method")
        with assert_raises(exceptions.FlexmockError, "cannot use at_most with at_least unset"):
            expectation.at_least().at_most()

    def test_at_most_cannot_be_specified_until_at_least_is_set(self):
        class Foo:
            def method(self):
                pass

        expectation = flexmock(Foo).should_receive("method")
        with assert_raises(exceptions.FlexmockError, "cannot use at_least with at_most unset"):
            expectation.at_most().at_least()



"""Tests for conditional assertions."""
# pylint: disable=missing-docstring,no-self-use,no-member
import functools

from flexmock import exceptions, flexmock
from tests.utils import assert_raises


class ConditionalAssertionsTestCase:
    def test_state_machine(self):
        class Radio:
            def __init__(self):
                self.is_on = False
                self.volume = 0

            def switch_on(self):
                self.is_on = True

            def switch_off(self):
                self.is_on = False

            def select_channel(self):
                return None

            def adjust_volume(self, num):
                self.volume = num

        radio = Radio()
        flexmock(radio)

        def radio_is_on():
            return radio.is_on

        radio.should_receive("select_channel").once().when(lambda: radio.is_on)
        radio.should_call("adjust_volume").once().with_args(5).when(radio_is_on)

        with assert_raises(
            exceptions.StateError,
            "select_channel expected to be called when lambda: radio.is_on is True",
        ):
            radio.select_channel()
        with assert_raises(
            exceptions.StateError, "adjust_volume expected to be called when radio_is_on is True"
        ):
            radio.adjust_volume(5)
        radio.is_on = True
        radio.select_channel()
        radio.adjust_volume(5)

    def test_when_parameter_should_be_callable(self):
        with assert_raises(exceptions.FlexmockError, "when() parameter must be callable"):
            flexmock().should_receive("something").when(1)

    def test_flexmock_should_not_blow_up_with_builtin_in_when(self):
        # It is not possible to get source for builtins. Flexmock should handle
        # this gracefully.
        mock = flexmock()
        mock.should_receive("something").when(functools.partial(lambda: False))
        with assert_raises(
            exceptions.StateError, "something expected to be called when condition is True"
        ):
            # Should not raise TypeError
            mock.something()



"""Flexmock tests that are not related to any particular feature."""
# pylint: disable=missing-docstring,no-self-use
from flexmock import Mock, flexmock
from flexmock._api import (
    ReturnValue,
    _format_args,
    _is_class_method,
    _is_static_method,
    _isproperty,
)
from tests.proxy import Proxy
from tests.some_module import DerivedClass, SomeClass


class CommonTestCase:
    def test_return_value_to_str(self):
        r_val = ReturnValue(value=1)
        assert str(r_val) == "1"

        r_val = ReturnValue(value=(1,))
        assert str(r_val) == "1"

        r_val = ReturnValue(value=(1, 2))
        assert str(r_val) == "(1, 2)"

        r_val = ReturnValue(value=1, raises=RuntimeError)
        assert str(r_val) == "<class 'RuntimeError'>(1)"

    def test_is_class_method(self):
        assert _is_class_method(SomeClass.class_method, "class_method") is True
        assert _is_class_method(SomeClass.static_method, "static_method") is False
        assert _is_class_method(SomeClass.instance_method, "instance_method") is False
        # Method names do no match
        assert _is_class_method(SomeClass.class_method, "other_method") is False

        some_class = SomeClass()
        assert _is_class_method(some_class.class_method, "class_method") is True
        assert _is_class_method(some_class.static_method, "static_method") is False
        assert _is_class_method(some_class.instance_method, "instance_method") is False

        assert _is_class_method(DerivedClass.class_method, "class_method") is True
        assert _is_class_method(DerivedClass.static_method, "static_method") is False
        assert _is_class_method(DerivedClass.instance_method, "instance_method") is False

        derived_class = DerivedClass()
        assert _is_class_method(derived_class.class_method, "class_method") is True
        assert _is_class_method(derived_class.static_method, "static_method") is False
        assert _is_class_method(derived_class.instance_method, "instance_method") is False

    def test_is_class_method_proxied(self):
        # pylint: disable=not-callable,invalid-name
        SomeClassProxy = Proxy(SomeClass)
        assert _is_class_method(SomeClassProxy.class_method, "class_method") is True
        assert _is_class_method(SomeClassProxy.static_method, "static_method") is False
        assert _is_class_method(SomeClassProxy.instance_method, "instance_method") is False

        some_class = SomeClassProxy()
        assert _is_class_method(some_class.class_method, "class_method") is True
        assert _is_class_method(some_class.static_method, "static_method") is False
        assert _is_class_method(some_class.instance_method, "instance_method") is False

        DerivedClassProxy = Proxy(DerivedClass)
        assert _is_class_method(DerivedClassProxy.class_method, "class_method") is True
        assert _is_class_method(DerivedClassProxy.static_method, "static_method") is False
        assert _is_class_method(DerivedClassProxy.instance_method, "instance_method") is False

        derived_class = DerivedClassProxy()
        assert _is_class_method(derived_class.class_method, "class_method") is True
        assert _is_class_method(derived_class.static_method, "static_method") is False
        assert _is_class_method(derived_class.instance_method, "instance_method") is False

    def test_is_static_method(self):
        assert _is_static_method(SomeClass, "class_method") is False
        assert _is_static_method(SomeClass, "static_method") is True
        assert _is_static_method(SomeClass, "instance_method") is False

        some_class = SomeClass()
        assert _is_static_method(some_class, "class_method") is False
        assert _is_static_method(some_class, "static_method") is True
        assert _is_static_method(some_class, "instance_method") is False

        assert _is_static_method(DerivedClass, "class_method") is False
        assert _is_static_method(DerivedClass, "static_method") is True
        assert _is_static_method(DerivedClass, "instance_method") is False

        derived_class = DerivedClass()
        assert _is_static_method(derived_class, "class_method") is False
        assert _is_static_method(derived_class, "static_method") is True
        assert _is_static_method(derived_class, "instance_method") is False

    def test_is_static_method_proxied(self):
        # pylint: disable=not-callable,invalid-name
        SomeClassProxy = Proxy(SomeClass)
        assert _is_static_method(SomeClassProxy, "class_method") is False
        assert _is_static_method(SomeClassProxy, "static_method") is True
        assert _is_static_method(SomeClassProxy, "instance_method") is False

        some_class = SomeClassProxy()
        assert _is_static_method(some_class, "class_method") is False
        assert _is_static_method(some_class, "static_method") is True
        assert _is_static_method(some_class, "instance_method") is False

        DerivedClassProxy = Proxy(DerivedClass)
        assert _is_static_method(DerivedClassProxy, "class_method") is False
        assert _is_static_method(DerivedClassProxy, "static_method") is True
        assert _is_static_method(DerivedClassProxy, "instance_method") is False

        derived_class = DerivedClassProxy()
        assert _is_static_method(derived_class, "class_method") is False
        assert _is_static_method(derived_class, "static_method") is True
        assert _is_static_method(derived_class, "instance_method") is False

    def test_isproperty(self):
        class BaseClass:
            @property
            def method1(self):
                pass

            def method2(self):
                pass

        class ChildClass(BaseClass):
            pass

        base_instance = BaseClass()
        child_instance = ChildClass()
        assert _isproperty(base_instance, "method1") is True
        assert _isproperty(base_instance, "method2") is False
        assert _isproperty(BaseClass, "method1") is True
        assert _isproperty(BaseClass, "method2") is False
        assert _isproperty(child_instance, "method1") is True
        assert _isproperty(child_instance, "method2") is False
        assert _isproperty(ChildClass, "method1") is True
        assert _isproperty(ChildClass, "method2") is False
        assert _isproperty(Mock(), "method2") is False

    def test_format_args_supports_tuples(self):
        formatted = _format_args("method", {"kargs": ((1, 2),), "kwargs": {}})
        assert formatted == "method((1, 2))"

    def test_flexmock_should_not_explode_on_unicode_formatting(self):
        formatted = _format_args("method", {"kargs": (chr(0x86C7),), "kwargs": {}})
        assert formatted == 'method("蛇")'

    def test_return_value_should_not_explode_on_unicode_values(self):
        return_value = ReturnValue(chr(0x86C7))
        assert f"{return_value}" == '"蛇"'
        return_value = ReturnValue((chr(0x86C7), chr(0x86C7)))
        assert f"{return_value}" == '("蛇", "蛇")'

    def test_flexmock_should_yield_self_when_iterated(self):
        class ClassNoIter:
            pass

        instance = flexmock(ClassNoIter)
        assert instance is list(instance)[0]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['flexmock']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'flexmock',
    'version': '0.11.3',
    'description': 'flexmock is a testing library for Python that makes it easy to create mocks, stubs and fakes.',
    'long_description': '<p align="center">\n  <img alt="banner" src="https://user-images.githubusercontent.com/25169984/138661460-969caf9e-8e88-4609-87c4-1a0ab9624ee4.png">\n</p>\n\n<p align="center"><strong>flexmock</strong> <em>- Mock, stub, and spy library for Python.</em></p>\n\n<p align="center">\n<a href="https://pypi.org/project/flexmock/">\n  <img src="https://img.shields.io/pypi/v/flexmock" alt="pypi">\n</a>\n<a href="https://github.com/flexmock/flexmock/actions/workflows/ci.yml">\n  <img src="https://github.com/flexmock/flexmock/actions/workflows/ci.yml/badge.svg" alt="ci">\n</a>\n<a href="https://flexmock.readthedocs.io/">\n  <img src="https://img.shields.io/readthedocs/flexmock" alt="documentation">\n</a>\n<a href="https://codecov.io/gh/flexmock/flexmock">\n  <img src="https://codecov.io/gh/flexmock/flexmock/branch/master/graph/badge.svg?token=wRgtiGxhiL" alt="codecov">\n</a>\n<a href="./LICENSE">\n  <img src="https://img.shields.io/pypi/l/flexmock" alt="license">\n</a>\n</p>\n\n---\n\nFlexmock is a testing library for Python that makes it easy to create mocks, stubs, and fakes.\n\n## Features\n\n- **Mock**: Easily create mock objects and make assertions about which methods or attributes were used and arguments they were called with.\n- **Spy**: Proxy calls to object\'s original methods or attributes and make assertions based on return values or call count.\n- **Fake**: Generate a fake objects to be used in your tests with ease.\n- **Stub**: Create stub objects which replace parts of existing objects and classes with just one call.\n- **No external dependencies**: Flexmock is lightweight and only uses Python standard library. There are no external dependencies.\n- **Simple and intuitive**: Declarations are structured to read more like English sentences than API calls, so they are easy to learn and use.\n- **Fully type annotated**: External API is fully type annotated so it works great with static analysis tools and editor auto-completion.\n- **Integrations with test runners**: Integrates seamlessly with all major test runners like unittest, doctest, and pytest.\n- **Python 3.6+ and PyPy3**: Extensively tested to work with latest Python versions.\n\n## Installation\n\nInstall with pip:\n\n```\npip install flexmock\n```\n\n## Examples\n\nFlexmock features smooth integration with pretty much every popular test runner, so no special setup is necessary. Simply importing flexmock into your test module is sufficient to get started with any of the following examples:\n\n```python\nfrom flexmock import flexmock\n```\n\n### Mocks\n\nAssertions take many flavors and flexmock has many different facilities to generate them:\n\n```python\n# Simplest is ensuring that a certain method is called\nflexmock(Train).should_receive("get_tickets").once()\n\n# Of course, it is also possible to provide a default return value\nflexmock(Train).should_receive("get_destination").and_return("Paris").once()\n\n# Or check that a method is called with specific arguments\nflexmock(Train).should_receive("set_destination").with_args("Seoul").at_least().twice()\n```\n\n### Spies\n\nInstead of mocking, there are also times when you want to execute the actual method and simply find out how many times it was called. Flexmock uses `should_call` to generate this sort of assertions instead of `should_receive`:\n\n```python\n# Verify that a method is called at most three times\nflexmock(Train).should_call("get_tickets").at_most().times(3)\n\n# Make sure that a method is never called with specific arguments\nflexmock(Train).should_call("set_destination").with_args("Helsinki").never()\n\n# More complex example with features like argument type and exception matching\nflexmock(Train).should_call("crash").with_args(str, int).and_raise(AttributeError).once()\n```\n\nSee more examples in the documentation.\n\n## Documentation\n\nUser guide, examples, and a full API reference is available at: https://flexmock.readthedocs.io\n\n## Contributing\n\nContributions are absolutely welcome and encouraged! See [CONTRIBUTING.md](https://github.com/flexmock/flexmock/blob/master/CONTRIBUTING.md) to get started.\n',
    'author': 'Slavek Kabrda',
    'author_email': None,
    'maintainer': 'Adarsh Krishnan',
    'maintainer_email': 'adarshk7@gmail.com',
    'url': 'https://github.com/flexmock/flexmock',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



"""Flexmock testing library for Python."""
from flexmock import _integrations  # pylint: disable=unused-import
from flexmock._api import Expectation, Mock, flexmock

__all__ = [
    "Expectation",
    "Mock",
    "flexmock",
]



"""Test TeamCity (PyCharm) integration."""
# pylint: disable=missing-docstring,no-self-use
import unittest

from teamcity.unittestpy import TeamcityTestRunner

from tests.features import FlexmockTestCase


class TestTeamCityTeardown(FlexmockTestCase, unittest.TestCase):
    """Test flexmock teardown works with TeamCity test runner (PyCharm)."""


if __name__ == "__main__":
    runner = TeamcityTestRunner()
    unittest.main(testRunner=runner)
